#  Copyright 2025 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

import asyncio
import json
from datetime import datetime
from itertools import islice
from typing import List, Dict, Tuple, Callable, Any, Union, Type, Set

from geojson import FeatureCollection, Feature
from geojson.geometry import Geometry, LineString, Point
from zepben.eas.client.eas_client import EasClient
from zepben.eas.client.study import Study, Result, GeoJsonOverlay
from zepben.ewb import PowerTransformer, ConductingEquipment, EnergyConsumer, AcLineSegment, \
    NetworkConsumerClient, PhaseCode, PowerElectronicsConnection, Feeder, PowerSystemResource, Location, \
    connect_with_token, NetworkTraceStep, Tracing, downstream, upstream, IncludedEnergizedContainers


with open("../config.json") as f:
    c = json.loads(f.read())


def chunk(it, size):
    it = iter(it)
    return iter(lambda: tuple(islice(it, size)), ())


async def main():
    # Only process feeders in the following zones
    zone_mrids = ["MTN"]
    print(f"Start time: {datetime.now()}")

    rpc_channel = connect_with_token(host=c["host"], access_token=c["access_token"], rpc_port=c["rpc_port"])
    client = NetworkConsumerClient(rpc_channel)
    hierarchy = (await client.get_network_hierarchy()).throw_on_error()
    substations = hierarchy.value.substations

    print(f"Collecting feeders from zones {', '.join(zone_mrids)}.")
    # Extract all the feeders to process
    feeder_mrids = []
    for zone_mrid in zone_mrids:
        if zone_mrid in substations:
            for feeder in substations[zone_mrid].feeders:
                feeder_mrids.append(feeder.mrid)

    print(f"Feeders to be processed: {', '.join(feeder_mrids)}")
    futures = []

    # Process the feeders in batches of 3, using asyncio, for performance
    batches = chunk(feeder_mrids, 3)
    for feeders in batches:
        all_traced_equipment = []
        transformer_to_suspect_end = dict()

        rpc_channel = connect_with_token(host=c["host"], access_token=c["access_token"], rpc_port=c["rpc_port"])
        print(f"Processing feeders {', '.join(feeders)}")
        for feeder_mrid in feeders:
            futures.append(asyncio.ensure_future(fetch_feeder_and_trace(feeder_mrid, rpc_channel)))

        for future in futures:
            tx_to_sus_end = await future
            if tx_to_sus_end:   # Empty if the feeder failed
                all_traced_equipment.extend(eq for (k, (count, sus_eq_list)) in tx_to_sus_end.items() for eq in sus_eq_list)
                transformer_to_suspect_end.update(tx_to_sus_end)

        print(f"Created Study for {len(feeder_mrids)} feeders")

        eas_client = EasClient(host=c["host"], port=c["rpc_port"], protocol="https", access_token=c["access_token"])

        print(f"Uploading Study for {', '.join(feeders)} ...")
        await upload_suspect_end_of_line_study(
            eas_client,
            all_traced_equipment,
            transformer_to_suspect_end,
            name=f"Suspect end of line {', '.join(feeders)}",
            description="Highlights every line that is downstream of transformer and ends without a consumer.",
            tags=["suspect_end_of_line", "-".join(zone_mrids)],
            styles=json.load(open("style_eol.json", "r"))
        )
        await eas_client.aclose()
        print(f"Uploaded Study")

    print(f"Finish time: {datetime.now()}")


def collect_eq_provider(collection: Set[ConductingEquipment]):

    async def collect_equipment(ps: NetworkTraceStep, _):
        collection.add(ps.path.to_equipment)

    return collect_equipment


async def get_downstream_eq(ce: ConductingEquipment) -> Set[ConductingEquipment]:
    equipment_set = set()

    await (
        Tracing.network_trace()
        .add_condition(downstream())
        .add_step_action(collect_eq_provider(equipment_set))
    ).run(start=ce, phases=PhaseCode.ABCN, can_stop_on_start_item=False)

    return equipment_set


async def fetch_feeder_and_trace(feeder_mrid: str, rpc_channel):
    print(f"Fetching Feeder {feeder_mrid}")
    client = NetworkConsumerClient(rpc_channel)

    result = (
        await client.get_equipment_container(
            mrid=feeder_mrid,
            expected_class=Feeder,
            include_energized_containers=IncludedEnergizedContainers.LV_FEEDERS
        )
    )
    if result.was_failure:
        print(f"Failed: {result.thrown}")
        return {}

    network = client.service

    print(f"Finished fetching Feeder {feeder_mrid}")
    print(f"Tracing downstream transformers for feeder {feeder_mrid}")
    transformer_to_eq: Dict[str, Set[ConductingEquipment]] = {}
    for io in (pt for pt in network.objects(PowerTransformer)):
        pt: PowerTransformer = io
        downstream_equipment = await get_downstream_eq(pt)
        transformer_to_eq[pt.mrid] = downstream_equipment

    print(f"Tracing suspect ends for feeder {feeder_mrid}")
    transformer_to_suspect_end = await get_transformer_to_suspect_end(transformer_to_eq)

    return transformer_to_suspect_end


async def get_transformer_to_suspect_end(
    transformer_to_eq: Dict[str, Set[ConductingEquipment]]
) -> Dict[str, Tuple[int, Set[ConductingEquipment]]]:

    transformer_to_suspect_end: Dict[str, (int, List[ConductingEquipment])] = {}
    for pt_mrid, eq_list in transformer_to_eq.items():
        single_terminal_junctions = [
            eq for eq in eq_list
            if not isinstance(eq, (EnergyConsumer, PowerElectronicsConnection))
               and len(list(eq.terminals)) == 1
        ]
        upstream_eq = set(
            await _get_upstream_eq_up_to_transformer(stj) for stj in single_terminal_junctions
        )

        transformer_to_suspect_end[pt_mrid] = (len(single_terminal_junctions), list(upstream_eq))

    return transformer_to_suspect_end


async def upload_suspect_end_of_line_study(
    eas_client: EasClient,
    pts: List[PowerTransformer],
    transformer_to_suspect_end: Dict[str, Tuple[int, List[ConductingEquipment]]],
    name: str,
    description: str,
    tags: List[str],
    styles: List
) -> None:

    class_to_properties = {
        EnergyConsumer: {
            "name": lambda ec: ec.name,
            "type": lambda x: "ec"
        },
        PowerTransformer: {
            "consumer_count": _suspect_end_count_from(transformer_to_suspect_end),
            "type": lambda x: "pt"
        },
        AcLineSegment: {"name": lambda ec: ec.name},
    }
    feature_collection = to_geojson_feature_collection(pts, class_to_properties)
    response = await eas_client.async_upload_study(
        Study(
            name=name,
            description=description,
            tags=tags,
            results=[
                Result(
                    name=name,
                    geo_json_overlay=GeoJsonOverlay(
                        data=feature_collection,
                        styles=[s['id'] for s in styles]
                    )
                )
            ],
            styles=styles
        )
    )
    print(f"Study response: {response}")


async def _get_upstream_eq_up_to_transformer(ce: ConductingEquipment) -> Set[ConductingEquipment]:
    eqs = set()

    await (
        Tracing.network_trace()
        .add_condition(upstream())
        .add_step_action(collect_eq_provider(eqs))
        .add_stop_condition(_is_transformer)
    ).run(start=ce, phases=PhaseCode.ABCN, can_stop_on_start_item=False)

    return eqs


async def _is_transformer(ps: NetworkTraceStep):
    return isinstance(ps.path.to_equipment, PowerTransformer)


def _suspect_end_count_from(pt_to_sus_end: Dict[str, Tuple[int, List[ConductingEquipment]]]):
    def fun(pt: PowerTransformer):
        count, suspect_eq = pt_to_sus_end.get(pt.mrid)
        return count if count else 0

    return fun


def to_geojson_feature_collection(
    psrs: List[PowerSystemResource],
    class_to_properties: Dict[Type, Dict[str, Callable[[Any], Any]]]
) -> FeatureCollection:

    features = []
    for psr in psrs:
        properties_map = class_to_properties.get(type(psr))

        if properties_map is not None:
            features.append(to_geojson_feature(psr, properties_map))

    return FeatureCollection(features)


def to_geojson_feature(
    psr: PowerSystemResource,
    property_map: Dict[str, Callable[[PowerSystemResource], Any]]
) -> Union[Feature, None]:

    geometry = to_geojson_geometry(psr.location)
    if geometry is None:
        return None

    properties = {k: f(psr) for (k, f) in property_map.items()}
    return Feature(psr.mrid, geometry, properties)


def to_geojson_geometry(location: Location) -> Union[Geometry, None]:
    points = list(location.points)
    if len(points) > 1:
        return LineString([(point.x_position, point.y_position) for point in points])
    elif len(points) == 1:
        return Point((points[0].x_position, points[0].y_position))
    else:
        return None


if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())



#  Copyright 2025 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.


import asyncio
import json

from zepben.ewb import (
    NetworkConsumerClient, PhaseCode, AcLineSegment, connect_with_token, EnergyConsumer,
    PowerTransformer, ConductingEquipment, Tracing, NetworkTraceStep, downstream, upstream,
    IncludedEnergizedContainers
)


async def main():
    with open("config.json") as f:
        c = json.loads(f.read())

    print("Connecting to Server")
    channel = connect_with_token(host=c["host"], access_token=c["access_token"], rpc_port=c["rpc_port"])

    client = NetworkConsumerClient(channel)
    result = (await client.get_network_hierarchy()).throw_on_error().result
    print("Connection Established")

    for feeder in result.feeders.values():
        if feeder.mrid != "WD24":
            continue
        print(f"\nFetching {feeder.mrid}")
        network = await get_feeder_network(channel, feeder.mrid)

        print("\nDownstream Trace Example..")
        # Get the count of customers per transformer
        for io in network.objects(PowerTransformer):
            customers = await get_downstream_customer_count(io, PhaseCode.ABCN)
            print(f"Transformer {io.mrid} has {customers} Energy Consumer(s)")

        print("\nUpstream Trace Example..")
        for ec in network.objects(EnergyConsumer):
            upstream_length = await get_upstream_length(ec, PhaseCode.ABCN)
            print(f"Energy Consumer {ec.mrid} --> Upstream Length: {upstream_length}")


async def get_feeder_network(channel, feeder_mrid):
    client = NetworkConsumerClient(channel)
    (await client.get_equipment_container(
        mrid=feeder_mrid,
        include_energized_containers=IncludedEnergizedContainers.LV_FEEDERS
    )).throw_on_error()
    return client.service


async def get_downstream_customer_count(ce: ConductingEquipment, phase_code: PhaseCode) -> int:
    customer_count = 0

    def collect_eq_in():
        async def add_eq(ps: NetworkTraceStep, _):
            nonlocal customer_count
            if isinstance(ps.path.to_equipment, EnergyConsumer):
                customer_count += 1
        return add_eq

    await (
        Tracing.network_trace()
        .add_condition(downstream())
        .add_step_action(collect_eq_in())
    ).run(start=ce, phases=phase_code)

    return customer_count


async def get_upstream_length(ce: ConductingEquipment, phases: PhaseCode) -> int:
    upstream_length = 0

    def collect_eq_in():
        async def add_eq(ps: NetworkTraceStep, _):
            nonlocal upstream_length
            equip = ps.path.to_equipment
            if isinstance(equip, AcLineSegment):
                if equip.length is not None:
                    upstream_length += equip.length
        return add_eq

    await (
        Tracing.network_trace()
        .add_condition(upstream())
        .add_step_action(collect_eq_in())
    ).run(start=ce, phases=phases)

    return upstream_length

if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())



#  Copyright 2025 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Example trace showing methods to traverse upstream of a given `IdentifiedObject` to find the first
occurrence of another specified `IdentifiedObject`
"""

import asyncio
import json

from zepben.ewb import (
    NetworkStateOperators, NetworkTraceActionType, NetworkTraceStep, StepContext,
    NetworkConsumerClient, ConductingEquipment, connect_with_token
)
from zepben.ewb import PowerTransformer, UsagePoint, Tracing, Switch, IncludedEnergizedContainers


with open("config.json") as f:
    c = json.loads(f.read())


def _trace(start_item, results, stop_condition):
    """Returns a `NetworkTrace` configured with our parameters"""

    def step_action(step: NetworkTraceStep, context: StepContext):
        if context.is_stopping:  # if the trace is stopping, we have found the equipment we're looking for
            results.append(step.path.to_equipment)

    state_operators = NetworkStateOperators.NORMAL

    return (
        Tracing.network_trace(
            network_state_operators=state_operators,
            action_step_type=NetworkTraceActionType.ALL_STEPS
        ).add_condition(state_operators.upstream())
        .add_stop_condition(stop_condition)
        .add_step_action(step_action)
        .add_start_item(start_item)
    )


async def main(mrid: str, feeder_mrid: str):
    channel = connect_with_token(host=c["host"], access_token=c["access_token"], rpc_port=c["rpc_port"])
    client = NetworkConsumerClient(channel)
    await client.get_equipment_container(feeder_mrid, include_energized_containers=IncludedEnergizedContainers.LV_FEEDERS)
    network = client.service

    try:
        usage_point = network.get(mrid, UsagePoint)
        # get the `ConductingEquipment` from the `UsagePoint`
        start_item = next(filter(lambda ce: isinstance(ce, ConductingEquipment), usage_point.equipment))
    except TypeError:
        start_item = network.get(mrid, ConductingEquipment)

    results = []

    # Get DSUB from which any given customer is supplied from using a basic upstream trace
    def dsub_stop_condition(step: NetworkTraceStep, _: StepContext):
        return isinstance(step.path.to_equipment, PowerTransformer)

    # Get Circuit Breaker from which any given customer is supplied from using a basic upstream trace
    # Uncomment stop condition below to use
    def circuit_breaker_stop_condition(step: NetworkTraceStep, _: StepContext):
        return isinstance(step.path.to_equipment, Switch)

    await _trace(
        start_item=start_item,
        results=results,
        stop_condition=dsub_stop_condition,
        # stop_condition=circuit_breaker_stop_condition,
    ).run()

    print(results)


if __name__ == "__main__":
    # EnergyConsumer: 50763684
    asyncio.run(main(mrid='4310990779', feeder_mrid='RW1292'))



#  Copyright 2023 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
import asyncio
import json

from geojson import Feature, LineString, FeatureCollection, Point
from zepben.eas import Study, Result, GeoJsonOverlay, EasClient
from zepben.ewb import AcLineSegment, EnergyConsumer, connect_with_token, NetworkConsumerClient, IncludedEnergizedContainers
# A study is a geographical visualisation of data that is drawn on top of the network.
# This data is typically the result of a load flow simulation.
# Each study may contain multiple results: different visualisations that the user may switch between.
# For example, the first result may display per-unit voltage data, while the second result highlights overloaded equipment.
# Two results are created in this example study: one makes a heatmap of energy consumers and the other highlights LV lines and displays their length.
# Both Evolve App Server and Energy Workbench must be running for this example.


with open("../config.json") as f:
    c = json.loads(f.read())


async def main():
    # Fetch network model from Energy Workbench's gRPC service (see ../connecting_to_grpc_service.py for examples on different connection functions)
    print("Connecting to EWB..")
    grpc_channel = connect_with_token(host=c["host"], access_token=c["access_token"], rpc_port=c["rpc_port"])
    feeder_mrid = "WD24"
    grpc_client = NetworkConsumerClient(grpc_channel)
    print("Connection established..")
    await grpc_client.get_equipment_container(feeder_mrid, include_energized_containers=IncludedEnergizedContainers.LV_FEEDERS)
    network = grpc_client.service

    print("Creating study..")
    # Make result that displays a heatmap of energy consumers.
    ec_geojson = []
    for ec in network.objects(EnergyConsumer):
        if ec.location is not None:
            coord = list(ec.location.points)[0]
            ec_feature = Feature(
                id=ec.mrid,
                geometry=Point((coord.x_position, coord.y_position))
            )
            ec_geojson.append(ec_feature)

    ec_result = Result(
        name="Energy Consumers",
        geo_json_overlay=GeoJsonOverlay(
            data=FeatureCollection(ec_geojson),
            styles=["ec-heatmap"]  # Select which Mapbox layers to show for this result
        )
    )

    # Make result that highlights LV lines. Each result is a named GeoJSON overlay.
    lv_lines_geojson = []
    for line in network.objects(AcLineSegment):
        if line.base_voltage_value <= 1000 and line.location is not None:
            line_feature = Feature(
                id=line.mrid,
                geometry=LineString([(p.x_position, p.y_position) for p in line.location.points]),
                properties={
                    "length": line.length  # Numeric and textual data may be added here. It will be displayed and formatted according to the style(s) used.
                }
            )
            lv_lines_geojson.append(line_feature)

    lv_lines_result = Result(
        name="LV Lines",
        geo_json_overlay=GeoJsonOverlay(
            data=FeatureCollection(lv_lines_geojson),
            styles=["lv-lines", "lv-lengths"]  # Select which Mapbox layers to show for this result
        )
    )

    # Create and upload the study.
    study = Study(
        name="Example Study",
        description="Example study with two results.",
        tags=["example"],  # Tags make it easy to search for studies in a large list of them.
        results=[ec_result, lv_lines_result],
        styles=json.load(open("style.json", "r"))  # This is the "layers" property of a Mapbox GL JS style.
        # Layers specify how features are rendered. For more information about layers, read https://docs.mapbox.com/mapbox-gl-js/style-spec/layers/.
        # Each layer may have an entry in the legend via the metadata["zb:legend"] field.
    )
    print("Study created..")
    print("Connecting to EAS..")
    eas_client = EasClient(host=c["host"], port=c["rpc_port"], protocol="https", access_token=c["access_token"])

    print("Connection established..")

    print("Uploading study...")
    response = await eas_client.async_upload_study(study)
    print(response)
    print("Study uploaded! Please check the Evolve Web App.")

    await eas_client.aclose()


if __name__ == "__main__":
    asyncio.run(main())



#  Copyright 2025 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Example trace showing method to traverse outwards from any given `IdentifiableObject` to the next
`Switch` object, and build a list of all contained equipment (isolate-able section)
"""

import asyncio
import json

from zepben.ewb import (
    NetworkStateOperators, NetworkTraceActionType, NetworkTraceStep, StepContext,
    NetworkConsumerClient, AcLineSegment, connect_with_token, stop_at_open, IncludedEnergizedContainers
)
from zepben.ewb import Tracing, Switch


async def main(conductor_mrid: str, feeder_mrid: str):
    with open("config.json") as f:
        c = json.loads(f.read())

    channel = connect_with_token(
        host=c["host"],
        access_token=c["access_token"],
        rpc_port=c["rpc_port"]
    )
    client = NetworkConsumerClient(channel)
    await client.get_equipment_container(
        feeder_mrid, include_energized_containers=IncludedEnergizedContainers.LV_FEEDERS
    )
    network = client.service

    hv_acls = network.get(conductor_mrid, AcLineSegment)

    found_equip = set()

    def queue_condition(step: NetworkTraceStep, context: StepContext, _, __):
        """Queue the next step unless it's a `Switch`"""
        return not isinstance(step.path.to_equipment, Switch)

    def step_action(step: NetworkTraceStep, context: StepContext):
        """Add to our list of equipment, and equipment stepped on during this trace"""
        found_equip.add(step.path.to_equipment.mrid)

    await (
        Tracing.network_trace(
            network_state_operators=NetworkStateOperators.NORMAL,
            action_step_type=NetworkTraceActionType.ALL_STEPS
        ).add_condition(stop_at_open())
        .add_queue_condition(queue_condition)
        .add_step_action(step_action)
        .add_start_item(hv_acls)
    ).run()

    # print a list of all mRID's for all equipment in the isolation area.
    print(found_equip)


if __name__ == "__main__":
    asyncio.run(main(conductor_mrid='50434998', feeder_mrid='RW1292'))



#  Copyright 2025 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
import asyncio
import json
import os
from dataclasses import dataclass
from typing import Optional
import pandas as pd

from zepben.ewb import NetworkConsumerClient, connect_with_token, ConductingEquipment, Feeder, IncludedEnergizedContainers

with open("./config.json") as f:
    c = json.loads(f.read())

"""
This is a basic example that shows how to export a CSV of all the conducting equipment in a feeder.
It will output one CSV per feeder in the network.
"""


async def connect():
    channel = connect_with_token(host=c["host"], rpc_port=c["rpc_port"], access_token=c["access_token"], ca_filename=c["ca_path"])
    network_client = NetworkConsumerClient(channel=channel)

    network_hierarchy = (await network_client.get_network_hierarchy()).throw_on_error().value
    print("Network hierarchy:")
    for gr in network_hierarchy.geographical_regions.values():
        print(f"- Geographical region: {gr.name}")
        for sgr in gr.sub_geographical_regions:
            print(f"  - Subgeographical region: {sgr.name}")
            for sub in sgr.substations:
                print(f"    - Zone Substation: {sub.name}")
                for fdr in sub.feeders:
                    print(f"      - Processing Feeder: {fdr.name}")
                    await process_nodes(fdr.mrid, channel)


@dataclass
class NetworkObject(object):
    id: str
    type: str
    feeder: str
    voltage: int
    dist_tx_id: Optional[str] = None
    dist_tx_name: Optional[str] = None


async def process_nodes(feeder_mrid: str, channel):
    print("Fetching from server ...")
    network_client = NetworkConsumerClient(channel=channel)
    network_service = network_client.service
    (await network_client.get_equipment_container(
        feeder_mrid,
        include_energized_containers=IncludedEnergizedContainers.LV_FEEDERS)
     ).throw_on_error()

    print("Processing equipment ...")
    feeder = network_service.get(feeder_mrid, Feeder)
    network_objects = []
    # Fetch all the high voltage conducting equipment from the Feeder
    for equip in feeder.equipment:
        if isinstance(equip, ConductingEquipment):
            no = NetworkObject(equip.mrid, type(equip).__name__, feeder_mrid, equip.base_voltage_value)
            network_objects.append(no)

    # Fetch all the low voltage conducting equipment from the LvFeeders supplied by this Feeder
    for lvf in feeder.normal_energized_lv_feeders:
        head = lvf.normal_head_terminal.conducting_equipment
        for equip in lvf.equipment:
            if isinstance(equip, ConductingEquipment):
                no = NetworkObject(equip.mrid, type(equip).__name__, feeder_mrid, equip.base_voltage_value, head.mrid, head.name)
                network_objects.append(no)

    print(f"Writing csvs/{feeder_mrid}_network_objects.csv")
    network_objects = pd.DataFrame(network_objects)
    os.makedirs("csvs", exist_ok=True)
    network_objects.to_csv(f"csvs/{feeder_mrid}_network_objects.csv", index=False)
    print(f"Finished processing {feeder_mrid}")


if __name__ == "__main__":
    asyncio.run(connect())



#  Copyright 2025 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
import asyncio
import logging
import pandapower as pp

from pp_creators.basic_creator import BasicPandaPowerNetworkCreator
from zepben.ewb import NetworkService, Terminal, EnergySource, Tracing

from zepben.examples.ieee_13_node_test_feeder import network

logger = logging.getLogger(__name__)


async def main():
    add_energy_source(network, network["br_650_t1"])
    await Tracing.set_direction().run(network)
    bbn_creator = BasicPandaPowerNetworkCreator(
        logger=logger,
        ec_load_provider=lambda ec: (5000, 0)  # Model each energy consumer with a 5kW nonreactive load
    )
    result = await bbn_creator.create(network)

    print(f"Translation successful: {result.was_successful}")
    print(result.network)
    print()

    print("bus table:")
    print(result.network["bus"])
    print()

    print("load table:")
    print(result.network["load"])
    print()

    print("ext_grid table:")
    print(result.network["ext_grid"])
    print()

    print("line table:")
    print(result.network["line"])
    print()

    print("trafo table:")
    print(result.network["trafo"])
    print()

    print("line_geodata table:")
    print(result.network["line_geodata"])
    print()

    print("Running load flow study...", end="")
    pp.runpp(result.network)
    print("done.")
    print()

    print(result.network)
    print()

    print("res_bus table:")
    print(result.network["res_bus"])
    print()

    print("res_line table:")
    print(result.network["res_line"])
    print()

    print("res_trafo table:")
    print(result.network["res_trafo"])
    print()

    print("res_ext_grid table:")
    print(result.network["res_ext_grid"])
    print()

    print("res_load table:")
    print(result.network["res_load"])
    print()


def add_energy_source(_network: NetworkService, connect_to_terminal: Terminal):
    bv = connect_to_terminal.conducting_equipment.base_voltage
    es_t = Terminal(phases=connect_to_terminal.phases)
    es = EnergySource(terminals=[es_t], base_voltage=bv)
    _network.add(es_t)
    _network.add(es)
    _network.connect_terminals(es_t, connect_to_terminal)


if __name__ == "__main__":
    asyncio.run(main())



#  Copyright 2025 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.


async def plaintext_connection():
    """
    Connects to an RPC server without TLS or authentication.
    This method should only be used in development and for demos.
    """
    from zepben.ewb import connect_insecure, NetworkConsumerClient
    async with connect_insecure("hostname", 1234) as insecure_channel:
        client = NetworkConsumerClient(insecure_channel)
        grpc_result = await client.get_network_hierarchy()
        print(grpc_result.result)


async def secure_connection():
    """
    Connects to an RPC server over TLS.
    No user/client credentials are used.
    """
    from zepben.ewb import connect_tls, NetworkConsumerClient
    async with connect_tls("hostname", 1234) as secure_channel:
        client = NetworkConsumerClient(secure_channel)
        grpc_result = await client.get_network_hierarchy()
        print(grpc_result.result)


async def secure_connection_with_user_credentials():
    """
    Connects to an RPC server over TLS with user credentials. The authentication config will be
    fetched from https://hostname/auth or https://hostname/ewb/auth by default, which includes
    the domain of the OAuth token provider.
    """
    from zepben.ewb import connect_with_password, NetworkConsumerClient
    async with connect_with_password(
        "client ID",
        "username",
        "password",
        "hostname",
        1234
    ) as secure_channel:
        client = NetworkConsumerClient(secure_channel)
        grpc_result = await client.get_network_hierarchy()
        print(grpc_result.result)

    # Specify authentication config explicitly
    from zepben.auth import AuthMethod
    async with connect_with_password(
        "client ID",
        "username",
        "password",
        "hostname",
        1234,
        audience="https://fake_audience/",
        issuer_domain="fake.issuer.domain",
        auth_method=AuthMethod.AUTH0
    ) as secure_channel:
        client = NetworkConsumerClient(secure_channel)
        grpc_result = await client.get_network_hierarchy()
        print(grpc_result.result)


async def secure_connection_with_client_credentials():
    """
    Connects to an RPC server over TLS with client credentials. The authentication config will be
    fetched from https://hostname/auth or https://hostname/ewb/auth by default, which includes the
    domain of the OAuth token provider.
    """
    from zepben.ewb import connect_with_secret, NetworkConsumerClient
    async with connect_with_secret(
        "client ID",
        "client secret",
        "hostname",
        1234
    ) as secure_channel:
        client = NetworkConsumerClient(secure_channel)
        grpc_result = await client.get_network_hierarchy()
        print(grpc_result.result)

    # Specify authentication config explicitly
    from zepben.auth import AuthMethod
    async with connect_with_secret(
        "client ID",
        "client secret",
        "hostname",
        1234,
        audience="https://fake_audience/",
        issuer_domain="fake.issuer.domain",
        auth_method=AuthMethod.AUTH0
    ) as secure_channel:
        client = NetworkConsumerClient(secure_channel)
        grpc_result = await client.get_network_hierarchy()
        print(grpc_result.result)


def connect_sync():
    """
    You may use `SyncNetworkConsumerClient` if you prefer not to use asyncio.
    The API calls are the same between `SyncNetworkConsumerClient` and `NetworkConsumerClient`.
    """
    from zepben.ewb import connect_insecure, SyncNetworkConsumerClient
    channel = connect_insecure("hostname", 1234)
    client = SyncNetworkConsumerClient(channel)
    grpc_result = client.get_network_hierarchy()
    print(grpc_result.result)


async def connect_using_token():
    import json
    from zepben.ewb import connect_with_token, NetworkConsumerClient

    with open("config.json") as f:
        c = json.load(f)

    print("Connecting to EWB..")
    channel = connect_with_token(
        host=c["host"],
        access_token=c["access_token"],
        rpc_port=c["rpc_port"]
    )
    client = NetworkConsumerClient(channel)
    print("Connection established..")
    print("Printing network hierarchy..")
    grpc_result = await client.get_network_hierarchy()
    print(grpc_result.result)


if __name__ == "__main__":
    import asyncio
    asyncio.run(connect_using_token())



#  Copyright 2025 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

import asyncio
import json
import os
import pandas as pd

from typing import Dict, Callable, List
from dataclasses import dataclass
from zepben.ewb import connect_with_token, NetworkConsumerClient, Feeder, Tracing, downstream, StepActionWithContextValue, \
    NetworkTraceStep, EnergyConsumer, StepContext, IdentifiedObject, Breaker, Fuse, PowerTransformer, TransformerFunctionKind, TreeNode, EquipmentTreeBuilder, \
    ConductingEquipment, NetworkTrace, upstream, IncludedEnergizedContainers


@dataclass
class EnergyConsumerDeviceHierarchy:
    energy_consumer_mrid: str
    lv_circuit_name: str
    upstream_switch_mrid: str
    lv_circuit_name: str
    upstream_switch_class: str
    distribution_power_transformer_mrid: str
    distribution_power_transformer_name: str
    regulator_mrid: str
    breaker_mrid: str
    feeder_mrid: str


def _get_client():
    with open('config.json') as f:
        config = json.load(f)

        # Connect to server
    channel = connect_with_token(**config)
    return NetworkConsumerClient(channel)


async def get_feeders(_client=None) -> Dict[str, Feeder]:
    _feeders = (await (_client or _get_client()).get_network_hierarchy()).result.feeders
    return _feeders


async def get_feeder_equipment(client: NetworkConsumerClient, feeder_mrid: str) -> None:
    """Get all objects under the feeder, including LV Feeders"""
    (await client.get_equipment_container(
        feeder_mrid,
        include_energized_containers=IncludedEnergizedContainers.LV_FEEDERS
    )).throw_on_error()


async def trace_from_energy_consumers(feeder_mrid: str, client=None):
    """
    Least efficient/the slowest
    Inefficient upstream tracing example.
    Trace upstream from every EnergyConsumer.
    """
    client = client or _get_client()
    await get_feeder_equipment(client, feeder_mrid)

    def _get_equipment_tree_trace(up_data: dict) -> NetworkTrace:
        def step_action(step: NetworkTraceStep, _: StepContext):
            to_equip: ConductingEquipment = step.path.to_equipment

            if isinstance(to_equip, Breaker):
                if not up_data.get('breaker'):
                    up_data['breaker'] = to_equip
            elif isinstance(to_equip, Fuse):
                if not up_data.get('upstream_switch'):
                    up_data['upstream_switch'] = to_equip
            elif isinstance(to_equip, PowerTransformer):
                if not up_data.get('distribution_power_transformer'):
                    up_data['distribution_power_transformer'] = to_equip
                elif not up_data.get('regulator') and to_equip.function == TransformerFunctionKind.voltageRegulator:
                    up_data['regulator'] = to_equip

        return (
            Tracing.network_trace()
            .add_condition(upstream())
            .add_step_action(step_action)
        )

    feeder = client.service.get(feeder_mrid, Feeder)

    energy_consumers = []
    for lvf in feeder.normal_energized_lv_feeders:
        for ce in lvf.equipment:
            if isinstance(ce, EnergyConsumer):
                up_data = {'feeder': feeder_mrid, 'energy_consumer_mrid': ce.mrid}

                # Trace upstream from EnergyConsumer.
                await _get_equipment_tree_trace(up_data).run(ce)
                energy_consumers.append(_build_row(up_data))

    write_csv(energy_consumers, feeder.mrid)


async def trace_from_feeder_downstream(feeder_mrid: str, client=None):
    """
    More memory use than `trace_from_feeder_context`, more efficient/faster than `trace_from_energy_consumers`
    Build an equipment tree of everything downstream of the feeder.
    Use the Equipment tree to recurse through parent equipment of all EC's and get the equipment we are interested in.
    """

    def process_leaf(up_data: dict, leaf: TreeNode):
        to_equip: IdentifiedObject = leaf.identified_object

        if isinstance(to_equip, Breaker):
            if not up_data.get('breaker'):
                up_data['breaker'] = to_equip
        elif isinstance(to_equip, Fuse):
            if not up_data.get('upstream_switch'):
                up_data['upstream_switch'] = to_equip
        elif isinstance(to_equip, PowerTransformer):
            if not up_data.get('distribution_power_transformer'):
                up_data['distribution_power_transformer'] = to_equip
            elif not up_data.get('regulator') and to_equip.function == TransformerFunctionKind.voltageRegulator:
                up_data['regulator'] = to_equip

    client = client or _get_client()
    await get_feeder_equipment(client, feeder_mrid)

    builder = EquipmentTreeBuilder()

    feeder = client.service.get(feeder_mrid, Feeder)
    await (
        Tracing.network_trace()
        .add_condition(downstream())
        .add_step_action(builder)
    ).run(getattr(feeder, 'normal_head_terminal'))

    energy_consumers = []

    for leaf in (l for l in builder.leaves if isinstance((ec := l.identified_object), EnergyConsumer)):
        ec_data = {'feeder': feeder.mrid, 'energy_consumer_mrid': ec.mrid}

        def _process(_leaf):
            process_leaf(ec_data, _leaf)
            if _leaf.parent:
                _process(_leaf.parent)

        _process(leaf)

        row = _build_row(ec_data)
        energy_consumers.append(row)

    write_csv(energy_consumers, feeder.mrid)


async def trace_from_feeder_context(feeder_mrid: str, client=None):
    """
    Most efficient/fastest.
    trace downstream from the feeder recording relevant information using `NetworkTrace` `StepContext`.
    """
    client = client or _get_client()
    # Get all objects under the feeder, including Substations and LV Feeders
    await get_feeder_equipment(client, feeder_mrid)

    energy_consumers = []

    feeder = client.service.get(feeder_mrid, Feeder)

    class StepActionWithContext(StepActionWithContextValue):
        def _apply(self, item: NetworkTraceStep, context: StepContext):
            if isinstance((ec := item.path.to_equipment), EnergyConsumer):
                nonlocal energy_consumers
                data = self.get_context_value(context)
                data.update({'feeder': feeder.mrid, 'energy_consumer_mrid': ec.mrid})

                row = _build_row(data)
                energy_consumers.append(row)

        def compute_next_value(self, next_item: NetworkTraceStep, current_item: NetworkTraceStep, current_value: Dict[str, IdentifiedObject]):
            data = dict(current_value)
            equip = next_item.path.to_equipment
            if isinstance(equip, Breaker):
                data['breaker'] = equip
            elif isinstance(equip, Fuse):
                data['upstream_switch'] = equip
            elif isinstance(equip, PowerTransformer):
                if equip.function == TransformerFunctionKind.distributionTransformer:
                    data['distribution_power_transformer'] = equip
                elif equip.function == TransformerFunctionKind.voltageRegulator:
                    data['regulator'] = equip
            return data

        def compute_initial_value(self, item: NetworkTraceStep):
            return {}

    await (
        Tracing.network_trace()
        .add_condition(downstream())
        .add_step_action(StepActionWithContext('key'))
    ).run(getattr(feeder, 'normal_head_terminal'))

    write_csv(energy_consumers, feeder.mrid)


def write_csv(energy_consumers: List[EnergyConsumerDeviceHierarchy], feeder_mrid: str):
    network_objects = pd.DataFrame(energy_consumers)
    os.makedirs("csvs", exist_ok=True)
    network_objects.to_csv(f"csvs/{feeder_mrid}_energy_consumers.csv", index=False)


class NullEquipment:
    """empty class to simplify code below in the case of an equipment not existing in that position of the network"""
    mrid = None
    name = None


def _build_row(up_data: dict[str, IdentifiedObject | str]) -> EnergyConsumerDeviceHierarchy:
    return EnergyConsumerDeviceHierarchy(
        energy_consumer_mrid=up_data['energy_consumer_mrid'],
        upstream_switch_mrid=(up_data.get('upstream_switch') or NullEquipment).mrid,
        lv_circuit_name=(up_data.get('upstream_switch') or NullEquipment).name,
        upstream_switch_class=type(up_data.get('upstream_switch')).__name__,
        distribution_power_transformer_mrid=(up_data.get('distribution_power_transformer') or NullEquipment).mrid,
        distribution_power_transformer_name=(up_data.get('distribution_power_transformer') or NullEquipment).name,
        regulator_mrid=(up_data.get('regulator') or NullEquipment).mrid,
        breaker_mrid=(up_data.get('breaker') or NullEquipment).mrid,
        feeder_mrid=up_data.get('feeder'),
    )


def process_feeders_sequentially():
    async def main_async(trace_type: Callable):
        """
        Fetch the equipment container from the given feeder and create a CSV with the relevant information.
        Differences between the functions passable as `trace_type` are documented in the relevant docstrings.

        `trace_type` must be one of the following:
            - `trace_from_energy_consumers`
            - `trace_from_energy_consumers_with_context`
            - `trace_from_feeder_downstream`

        """
        from tqdm import tqdm
        client = _get_client()
        feeders = ["<FEEDER_ID>"]
        # feeders = list(await get_feeders(client)) # Uncomment to process all feeders
        for _feeder in tqdm(feeders):
            await trace_type(_feeder, client)

    # Uncomment to run other trace functions
    asyncio.run(main_async(trace_from_feeder_context))
    # asyncio.run(main_async(trace_from_feeder_downstream))
    # asyncio.run(main_async(trace_from_energy_consumers))


def process_feeders_concurrently():
    def multi_proc(_feeder):
        # Uncomment to run other trace functions
        asyncio.run(trace_from_feeder_context(_feeder))
        # asyncio.run(trace_from_feeder_downstream(_feeder))
        # asyncio.run(trace_from_energy_consumers(_feeder))

    # Get a list of feeders before entering main compute section of script.
    feeders = list(asyncio.run(get_feeders()))

    from tqdm.contrib.concurrent import process_map
    process_map(multi_proc, feeders, max_workers=int(os.cpu_count() / 2))


if __name__ == "__main__":
    process_feeders_sequentially()
    # process_feeders_concurrently()  # Uncomment and comment sequentially above to multi-process, note this is resource intensive and may cause issues.



#  Copyright 2025 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
from zepben.ewb import (
    EnergySource, AcLineSegment, Fuse, PowerTransformer, Breaker, EnergyConsumer, NetworkService,
    Terminal, connected_equipment, ConductingEquipment, PhaseCode, connected_terminals,
    ConnectivityResult
)

def build_network() -> NetworkService:
    """
    This function will return a network model resembling the below.

       source     consumer
         |           |
        line       line
         |           |
        fuse      breaker
         |           |
    transformer     fuse
         |           |
         +-----------+
    """

    # We create the objects, and their Terminals
    _es = EnergySource(mrid="es", terminals=[
        Terminal(mrid="es_t")
    ])

    _hv_line = AcLineSegment(mrid="hv_line", terminals=[
        Terminal(mrid="hv_line_t1"),
        Terminal(mrid="hv_line_t2")
    ])

    _hv_fuse = Fuse(mrid="hv_fuse", terminals=[
        Terminal(mrid="hv_fuse_t1"),
        Terminal(mrid="hv_fuse_t2")
    ])

    _tx = PowerTransformer(mrid="tx", terminals=[
        Terminal(mrid="tx_t1"),
        Terminal(mrid="tx_t2", phases=PhaseCode.ABCN)
    ])

    _lv_fuse = Fuse(mrid="lv_fuse", terminals=[
        Terminal(mrid="lv_fuse_t1", phases=PhaseCode.ABCN),
        Terminal(mrid="lv_fuse_t2", phases=PhaseCode.ABCN)
    ])

    _breaker = Breaker(mrid="breaker", terminals=[
        Terminal(mrid="breaker_t1", phases=PhaseCode.ABCN),
        Terminal(mrid="breaker_t2", phases=PhaseCode.BN)
    ])

    _lv_line = AcLineSegment(mrid="lv_line", terminals=[
        Terminal(mrid="lv_line_t1", phases=PhaseCode.BN),
        Terminal(mrid="lv_line_t2", phases=PhaseCode.BN)
    ])

    _ec = EnergyConsumer(mrid="ec", terminals=[
        Terminal(mrid="ec_t", phases=PhaseCode.BN)
    ])

    # Now we add the objects and their terminals to the network
    network = NetworkService()
    for io in (_es, _hv_line, _hv_fuse, _tx, _lv_fuse, _breaker, _lv_line, _ec):
        network.add(io)  # add the object
        for terminal in io.terminals:  # iterate over Terminals
            network.add(terminal)  # add them too

    # Power grids aren't much use if the equipment in them isn't connected to anything,
    # Lets connect them.
    network.connect_terminals(network['es_t'], network['hv_line_t1'])
    network.connect_terminals(network['hv_line_t2'], network['hv_fuse_t1'])
    network.connect_terminals(network['hv_fuse_t2'], network['tx_t1'])
    network.connect_terminals(network['tx_t2'], network['lv_fuse_t1'])
    network.connect_terminals(network['lv_fuse_t2'], network['breaker_t1'])
    network.connect_terminals(network['breaker_t2'], network['lv_line_t1'])
    network.connect_terminals(network['lv_line_t2'], network['ec_t'])

    return network


def fancy_print_connectivity_result(_connectivity_result: ConnectivityResult):
    print(f"\t{_connectivity_result.from_terminal} to {_connectivity_result.to_terminal}")

    terminal_str_len = len(str(_connectivity_result.from_terminal))
    for core_path in _connectivity_result.nominal_phase_paths:
        print(f"\t{core_path.from_phase.name:>{terminal_str_len}}----{core_path.to_phase.name}")


def fancy_print_connected_equipment(equipment: ConductingEquipment, phases=None):
    if phases:
        print(f"Connectivity results for {equipment} on phases {phases}:")
    else:
        print(f"Connectivity results for {equipment}:")
    for _connectivity_result in connected_equipment(equipment, phases):
        fancy_print_connectivity_result(_connectivity_result)
    print()


if __name__ == '__main__':
    # This example explores how to examine the immediate connectivity of equipment.
    # We will build a simple, linear network to examine:
    n = build_network()

    # Get references to the ConductingEquipment we are interested in from the network
    tx = n['tx']
    breaker = n['breaker']
    lv_fuse_t2 = n['lv_fuse_t2']

    # connected_equipment(equipment, phases) will get all connections between equipment cores
    # matching one of the requested phases. The connected equipment does not need to connect
    # via all specified phases to appear in the list of connectivity results.
    fancy_print_connected_equipment(tx)
    fancy_print_connected_equipment(tx, phases=PhaseCode.N)
    fancy_print_connected_equipment(breaker, phases=PhaseCode.BC)

    # connected_terminals is essentially connected_equipment where only one terminal is considered.
    print(f"Connectivity results for terminal {lv_fuse_t2} on phases {PhaseCode.ACN}:")
    for connectivity_result in connected_terminals(lv_fuse_t2, PhaseCode.ACN):
        fancy_print_connectivity_result(connectivity_result)



#  Copyright 2025 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
import asyncio
import json

from zepben.ewb import connect_with_token, NetworkConsumerClient

with open("config.json") as f:
    c = json.loads(f.read())


async def main():
    # See connecting_to_grpc_service.py for examples of each connect function
    print("Connecting to EWB..")
    channel = connect_with_token(host=c["host"], access_token=c["access_token"], rpc_port=c["rpc_port"])
    client = NetworkConsumerClient(channel)
    print("Connection established..")
    # Fetch network hierarchy
    network_hierarchy = await client.get_network_hierarchy()

    print("Network hierarchy:")
    for gr in network_hierarchy.result.geographical_regions.values():
        print(f"- GeographicalRegion mRID: {gr.mrid} name: {gr.name}")
        for sgr in gr.sub_geographical_regions:
            print(f"  - SubgeographicalRegion mRID: {sgr.mrid} name: {sgr.name}")
            for sub in sgr.substations:
                print(f"    - Substation mRID: {sub.mrid} name: {sub.name}")
                for fdr in sub.feeders:
                    print(f"      - Feeder mRID: {fdr.mrid} name: {fdr.name}")


if __name__ == "__main__":
    asyncio.run(main())



#  Copyright 2025 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
import json
from typing import Dict

from zepben.eas.client.eas_client import EasClient

import logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s: %(message)s')
logger = logging.getLogger()


# Default config dir is where the sample_auth_config sits.
def get_config_dir(argv):
    return argv[1] if len(argv) > 1 else "."


def read_json_config(config_file_path: str) -> Dict:
    file = open(config_file_path)
    config_dict = json.load(file)
    file.close()
    return config_dict


def get_client(config_dir):
    # Change sample_auth_config.json to any other file name
    auth_config = read_json_config(f"{config_dir}/sample_auth_config.json")

    return EasClient(
        host=auth_config["eas_server"]["host"],
        port=auth_config["eas_server"]["port"],
        protocol=auth_config["eas_server"]["protocol"],
        access_token=auth_config["eas_server"]["access_token"],
        verify_certificate=auth_config["eas_server"].get("verify_certificate", True),
        ca_filename=auth_config["eas_server"].get("ca_filename")
    )



#  Copyright 2025 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from zepben.ewb import NetworkHierarchy, GeographicalRegion, SubGeographicalRegion, Feeder, Substation, Loop, Circuit

if __name__ == '__main__':

    # A network hierarchy describes the high-level hierarchy of the network.

    fdr1 = Feeder(name="Sydney feeder 1")
    fdr2 = Feeder(name="Sydney feeder 2")
    fdr3 = Feeder(name="Sydney feeder 3")
    fdr4 = Feeder(name="Newcastle feeder 1")
    fdr5 = Feeder(name="Newcastle feeder 2")
    fdr6 = Feeder(name="Newcastle feeder 3")

    sub1 = Substation(name="Sydney substation 1")
    sub2 = Substation(name="Sydney substation 2", normal_energized_feeders=[fdr1, fdr2, fdr3])
    sub3 = Substation(name="Newcastle substation", normal_energized_feeders=[fdr4, fdr5, fdr6])

    circuit_sydney = Circuit(end_substations=[sub1, sub2])
    loop_sydney = Loop(circuits=[circuit_sydney], substations=[sub1], energizing_substations=[sub2])
    sgr_sydney = SubGeographicalRegion(name="Sydney", substations=[sub1, sub2])
    sgr_newcastle = SubGeographicalRegion(name="Newcastle", substations=[sub3])

    gr_nsw = GeographicalRegion(name="New South Wales", sub_geographical_regions=[sgr_sydney, sgr_newcastle])

    network_hierarchy = NetworkHierarchy(
        geographical_regions={gr_nsw.mrid: gr_nsw},
        sub_geographical_regions={sgr.mrid: sgr for sgr in (sgr_sydney, sgr_newcastle)},
        substations={sub.mrid for sub in (sub1, sub2, sub3)},
        feeders={fdr.mrid: fdr for fdr in (fdr1, fdr2, fdr3, fdr4, fdr5, fdr6)},
        circuits={circuit_sydney.mrid: circuit_sydney},
        loops={loop_sydney.mrid: loop_sydney}
    )

    print("Network hierarchy:")
    for gr in network_hierarchy.geographical_regions.values():
        print(f"- {gr.name}")
        for sgr in gr.sub_geographical_regions:
            print(f"  - {sgr.name}")
            for sub in sgr.substations:
                print(f"    - {sub.name}")
                for fdr in sub.feeders:
                    print(f"      - {fdr.name}")


#  Copyright 2025 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
import json
from datetime import datetime

from zepben.eas.client.opendss import OpenDssConfig
from zepben.eas.client.work_package import GeneratorConfig, ModelConfig, FeederScenarioAllocationStrategy, SolveConfig, RawResultsConfig, \
    MeterPlacementConfig, SwitchMeterPlacementConfig, SwitchClass
from zepben.eas import EasClient, TimePeriod, FixedTime
from time import sleep
import requests


with open("config.json") as f:
    c = json.loads(f.read())


def wait_for_export(eas_client: EasClient, model_id: int):
    # Wait for OpenDss model export to complete
    wait_limit_seconds = 3000
    step_seconds = 2
    total = 0
    print(f"Waiting for model generation ({wait_limit_seconds} seconds) ", end='')
    # Retrieve the model information for the model we just requested
    model = eas_client.get_opendss_model(model_id)
    while model["state"] != "COMPLETED":
        model = eas_client.get_opendss_model(model_id)
        print(".", end='')
        sleep(step_seconds)
        total += step_seconds
        if total > wait_limit_seconds:
            raise TimeoutError("Timed out waiting for model export to complete.")


def download_generated_model(eas_client: EasClient, output_file_name: str, model_id: int):
    url = eas_client.get_opendss_model_download_url(model_id)
    if url == f'Model with id {model_id} is still being created':
        print(url)
        print("Download failed.")
        return

    print(f"\nURL (30 second expiry): {url}", )

    file_name = f"{output_file_name}-{model_id}.zip"
    print(f"Downloading model zip to: {file_name}")

    try:
        with open(file_name, mode="wb") as file:
            file.write(requests.get(url).content)
        print("Download complete.")
    except Exception as error:
        print("Download failed.")
        print(error)


def test_open_dss_export(export_file_name: str):
    eas_client = EasClient(
        host=c["host"],
        port=c["rpc_port"],
        access_token=c["access_token"]
    )

    # Run an opendss export
    print("Sending OpenDss model export request to EAS")

    response = eas_client.run_opendss_export(
        OpenDssConfig(
            scenario="base",
            year=2025,
            feeder="<FEEDER_MRID>",
            load_time=TimePeriod(
                start_time=datetime.fromisoformat("2024-04-01T00:00"),
                end_time=datetime.fromisoformat("2025-04-01T00:00")
            ),
            # For fixed time export example, pass load_time a FixedTime object
            #load_time=FixedTime(
            #    time=datetime.fromisoformat("2024-04-01T00:00")
            #),
            generator_config=GeneratorConfig(
                model=ModelConfig(
                    meter_placement_config=MeterPlacementConfig(
                        feeder_head=True,
                        dist_transformers=True,
                        # Include meters for any switch that has a name that starts with 'LV Circuit Head' and is a Fuse or Disconnector
                        switch_meter_placement_configs=[SwitchMeterPlacementConfig(
                            meter_switch_class=SwitchClass.DISCONNECTOR,
                            name_pattern="LV Circuit Head.*"
                        ), SwitchMeterPlacementConfig(
                            meter_switch_class=SwitchClass.FUSE,
                            name_pattern="LV Circuit Head.*"
                        )]
                    ),
                    vmax_pu=1.2,
                    vmin_pu=0.8,
                    p_factor_base_exports=-1,
                    p_factor_base_imports=1,
                    p_factor_forecast_pv=0.98,
                    fix_single_phase_loads=True,
                    max_single_phase_load=15000.0,
                    max_load_service_line_ratio=1.0,
                    max_load_lv_line_ratio=2.0,
                    max_load_tx_ratio=2.0,
                    max_gen_tx_ratio=4.0,
                    fix_overloading_consumers=True,
                    fix_undersized_service_lines=True,
                    feeder_scenario_allocation_strategy=FeederScenarioAllocationStrategy.ADDITIVE,
                    closed_loop_v_reg_enabled=True,
                    closed_loop_v_reg_set_point=0.9825,
                    seed=123,
                ),
                solve=SolveConfig(step_size_minutes=30.0),
                raw_results=RawResultsConfig(True, True, True, True, True)
            ),
            model_name=export_file_name,
            is_public=True
        )
    )
    print(f"Raw 'run_opendss_export' response: '{response}'")
    model_id = response["data"]["createOpenDssModel"]
    print(f"New OpenDss model export id: {model_id}")

    try:
        wait_for_export(eas_client, int(model_id))

        # Request a download URL from EAS and download to a local file
        download_generated_model(eas_client, export_file_name, int(model_id))

    except TimeoutError:
        print("\nERROR: Timed out waiting for model export to complete.")

    eas_client.close()


if __name__ == "__main__":
    test_open_dss_export("test_export-model")



#  Copyright 2025 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

import asyncio
import json
from typing import List, Set

from zepben.ewb import (
    Feeder, PowerTransformer, Switch, Tracing, NetworkConsumerClient, Terminal, BusbarSection,
    ConductingEquipment, Breaker, EquipmentContainer, StepContext, NetworkTraceStep, connect_with_token,
    IncludedEnergizedContainers, IncludedEnergizingContainers
)

"""
Primary question to answer/example for:
1. How to access the CIM model? Show examples of how the static/design and dynamic/current states
   of the network model is typically accessed by software developers?
    a. Can Zepben model be updated with dynamic ADMS state information
    b. if we have current state of network (dynamically updated from ADMS), can we query the model
       to find all current connected HV feeders in a voltage bus? (VVC Example)
    c. How fast can we retrieve a model (dynamic sate) from CIM for near real time applications? 
       (VVC Example)
2. Show how the static and dynamic states of the network model is used by applications.
"""


async def run_simple(client: NetworkConsumerClient):
    for heading, function in (
            ('FETCH ZONE FEEDERS', fetch_zone_feeders),
            ("SPUR ISOLATION", isolate_spur_current),
            ("ZONE BUS TRACE", zone_bus_trace)
    ):
        print(f"\n\n######################\n# {heading} #\n######################\n\n")
        await function(client)


async def fetch_zone_feeders(client: NetworkConsumerClient):
    print("fetching network hierarchy...")
    hierarchy = (await client.get_network_hierarchy()).throw_on_error().result
    print("hierarchy fetched, fetching BAS feeders...")
    substation = hierarchy.substations["BAS"]
    for feeder in substation.feeders:
        print(f"   {feeder.mrid}...")
        await client.get_equipment_container(
            feeder.mrid,
            Feeder,
            include_energizing_containers=IncludedEnergizingContainers.FEEDERS,
            include_energized_containers=IncludedEnergizedContainers.FEEDERS
        )
    print("CPM feeders fetched.")


async def isolate_spur_current(client: NetworkConsumerClient):
    feeder: Feeder = client.service.get("BAS022", Feeder)
    switch: Switch = client.service.get("171143844", Switch)
    tx: PowerTransformer = client.service.get("171143830", PowerTransformer)

    log_spur("original:", switch, tx)

    clear_feeders({feeder})

    print(f"setting switch open...")

    log_spur("cleaned:", switch, tx)

    # Change the open state of the switch.
    switch.set_open(True)

    await recalculate_feeders({feeder})

    log_spur("currently open:", switch, tx)


def log_spur(desc: str, switch: Switch, tx: PowerTransformer):
    print(f"==========================\n{desc}"
          f"\n   {str(switch)}: is_normally_open={switch.is_normally_open()}, "
          f"is_open={switch.is_open()}, "
          f"normal_feeders={[it.mrid for it in switch.normal_feeders]}, "
          f"current_feeders={[it.mrid for it in switch.current_feeders]}"
          f"\n   {str(tx)}: nominal_phases={[it.phases.name for it in tx.terminals]}, "
          f"normal_phases={[it.normal_phases.as_phase_code().name for it in tx.terminals]}, "
          f"current_phases={[it.current_phases.as_phase_code().name for it in tx.terminals]}, "
          f"normal_feeders={[it.mrid for it in tx.normal_feeders]}, "
          f"current_feeders={[it.mrid for it in tx.current_feeders]}"
          "\n=========================="
    )


async def zone_bus_trace(client: NetworkConsumerClient):
    feeder_head_terminals = []
    feeder_heads = []
    feeder_head_other_terms = []
    for feeder in client.service.objects(Feeder):
        if (head_terminal := feeder.normal_head_terminal) and head_terminal.conducting_equipment:
            feeder_head_terminals.append(head_terminal)
            feeder_heads.append(head_terminal.conducting_equipment)
            feeder_head_other_terms.extend(head_terminal.other_terminals())

    print(f"creating bus for {[feeder.mrid for it in feeder_heads for feeder in it.normal_feeders]}...")
    # There is no subtrans in the model we pulled down so create a zone bus for all the feeders.
    bus = BusbarSection()
    bus_terminal = Terminal()
    bus.add_terminal(bus_terminal)

    for it in feeder_head_other_terms:
        print(f"   connecting {[feeder.mrid for feeder in it.conducting_equipment.normal_feeders]} to bus...")
        # disconnect the terminal from its energy source and move it to the bus
        client.service.disconnect(it)
        client.service.connect_terminals(bus_terminal, it)

    print("bus created")
    await log_bus("original:", bus, feeder_heads)

    print("Setting BAS022 and BAS023 to currently open")
    client.service.get("15416104", Breaker).set_open(True)  # BAS022 Breaker
    client.service.get("15416128", Breaker).set_open(True)  # BAS023 Breaker

    print("Setting BAS032 and BAS033 to normally open")
    client.service.get("177750988", Breaker).set_normally_open(True)  # BAS032 Breaker
    client.service.get("177750866", Breaker).set_normally_open(True)  # BAS033 Breaker

    await log_bus("opened:", bus, feeder_heads)


async def log_bus(desc: str, bus: ConductingEquipment, feeder_heads: List[ConductingEquipment]):
    print(f"==========================\n{desc}")

    # we run a trace on the assumption that the real model may have more equipment between the bus and the feeder heads.
    # e.g. other minor busbars or ac line segments
    open_heads: List[ConductingEquipment] = []
    closed_heads: List[ConductingEquipment] = []
    normally_open_heads: List[ConductingEquipment] = []
    normally_closed_heads: List[ConductingEquipment] = []

    # stop at all feeder heads
    def stop_on_feeder_heads(step: NetworkTraceStep, _: StepContext) -> bool:
        return isinstance(step.path.to_equipment, Switch) and step.path.to_equipment in feeder_heads

    # stop at transformers to prevent tracing out of this zone into others
    def stop_on_transformers(step: NetworkTraceStep, _: StepContext) -> bool:
        return isinstance(step.path.to_equipment, PowerTransformer)

    # sort feeder heads based on state
    def sort_feeder_heads(step: NetworkTraceStep, _: StepContext) -> None:
        to_equipment = step.path.to_equipment
        if isinstance(to_equipment, Switch):
            if to_equipment.is_open():
                open_heads.append(to_equipment)
            else:
                closed_heads.append(to_equipment)
            if to_equipment.is_normally_open():
                normally_open_heads.append(to_equipment)
            else:
                normally_closed_heads.append(to_equipment)

    await (
        Tracing.network_trace()
        .add_stop_condition(stop_on_feeder_heads)
        .add_stop_condition(stop_on_transformers)
        .if_stopping(sort_feeder_heads)
    ).run(start=bus)

    print(
        f"   disconnected feeders: {[feeder.mrid for it in open_heads for feeder in it.normal_feeders]}"
        f"\n   connected feeders: {[feeder.mrid for it in closed_heads for feeder in it.normal_feeders]}"
        f"\n   normally disconnected feeders: {[feeder.mrid for it in normally_open_heads for feeder in it.normal_feeders]}"
        f"\n   normally connected feeders: {[feeder.mrid for it in normally_closed_heads for feeder in it.normal_feeders]}"
        "\n=========================="
    )


async def run_swap_feeder(client: NetworkConsumerClient):
    open_point_id = "13953031"
    isolation_point_id = "13952991"

    # noinspection PyTypeChecker
    open_point: Switch = (await client.get_identified_object(open_point_id)).throw_on_error().value
    # noinspection PyTypeChecker
    isolation_point: Switch = (await client.get_identified_object(isolation_point_id)).throw_on_error().value

    feeder_ids = {it.to_mrid for it in client.service.get_unresolved_references_from(open_point.mrid) if it.resolver.to_class == EquipmentContainer}

    # get the feeders on both sides of the open point.
    print(f"fetching feeders...")
    for feeder in feeder_ids:
        print(f"   {feeder}...")
        (await client.get_equipment_container(feeder)).throw_on_error()
    print(f"done.")

    feeders = set(open_point.normal_feeders)
    log_txs(f"original:", feeders)

    clear_feeders(feeders)

    print(f"setting switches to move feeders...")

    open_point.set_open(False)
    isolation_point.set_open(True)

    print(f"switches updated.")

    await recalculate_feeders(feeders)

    log_txs(f"swapped:", feeders)


def clear_feeders(feeders: Set[Feeder]):
    # remove the phases and feeders to show the difference in open/normal state
    for feeder in feeders:
        print(f"removing phases from {feeder.mrid}...")
        Tracing.remove_phases().run(start=feeder.normal_head_terminal)

        print(f"phases removed, removing equipment...")
        for equip in feeder.equipment:
            equip.clear_containers()
            equip.clear_current_containers()
        feeder.clear_equipment()
        feeder.clear_current_equipment()
        print(f"equipment removed.")


async def recalculate_feeders(feeders: Set[Feeder]):
    # recalculate the phases and feeders with the new switch state.
    for feeder in feeders:
        print(f"assigning phases to {feeder.mrid}...")
        await Tracing.set_phases().run_with_terminal(feeder.normal_head_terminal)
        print(f"phases assigned, assigning equipment...")
        await Tracing.assign_equipment_to_feeders().run_feeder(feeder)
        print(f"equipment assigned.")


def log_txs(desc: str, feeders: Set[Feeder]):
    print()
    print(desc)

    for feeder in feeders:
        print(f"   {feeder.mrid} txs: {sorted([tx.name for tx in feeder.current_equipment if isinstance(tx, PowerTransformer)])}")

    print()


async def main():

    # noinspection PyTypeChecker
    with open('config.json') as f:
        config = json.load(f)
    async with connect_with_token(**config) as secure_channel:
        await run_simple(NetworkConsumerClient(secure_channel))
        await run_swap_feeder(NetworkConsumerClient(secure_channel))


if __name__ == "__main__":
    asyncio.run(main())



#  Copyright 2025 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
from zepben.ewb.services.common.resolver import per_length_impedance
from zepben.ewb import (
    NetworkService, AcLineSegment, PerLengthSequenceImpedance, Switch, Breaker,
    ConductingEquipment, NameType, Meter, EnergySource, Terminal
)

# A `NetworkService` is a mutable node breaker network model that implements a subset of
# IEC61968 and IEC61970 CIM classes. It is essentially a collection of `IdentifiedObject`s,
# and they may be added and removed as desired.
network = NetworkService()

print("\n##################\n# ADDING OBJECTS #\n##################\n")
# We start by adding a line segment and a breaker to the network model.
line = AcLineSegment(mrid="acls_123")
breaker = Breaker(mrid="b_456")

print(f"{line} added? {network.add(line)}")
print(f"{breaker} added? {network.add(breaker)}")

# Objects with duplicate mRIDs are not added.
invalid = EnergySource(mrid="acls_123")
print(f"{invalid} added? {network.add(invalid)}")


print("\n####################\n# QUERYING OBJECTS #\n\n####################\n")
# Use the `get` method to query the network model for an object with the specified mRID.
print(f"Identified object with mrid acls_123: {network.get('acls_123')}")
print(f"Identified object with mrid b_456: {network.get('b_456')}")

# A `KeyError` is raised if no object with the specified mRID is in the network model.
try:
    network.get("not_in_network")
except KeyError as error:
    print(error)

# Narrow the desired type with the second parameter. In makes the intent clearer, and
# lets IDEs lint and autocomplete according to the requested type.
print(f"Switch with mrid b_456 is open? {network.get('b_456', Switch).is_open()}")

# A `TypeError` is raised if the object exists in the network model, but is not the correct type.
try:
    network.get("acls_123", Switch)
except TypeError as error:
    print(error)

print("\n##################\n# QUERYING TYPES #\n##################\n")
# You may use the `objects` method to iterate over all objects that inherit a specified type.
# Because the breaker is the only object in the network model that inherits from the `Switch`
# class, this will print "Switch: Breaker{b_456}".
for switch in network.objects(Switch):
    print(f"Switch: {switch}")

# However, both the line and the breaker inherit from `ConductingEquipment`.
# The following line prints "Conducting equipment: AcLineSegment{acls_123}"
# and "Conducting equipment: Breaker{b_456}".
for conducting_equipment in network.objects(ConductingEquipment):
    print(f"Conducting equipment: {conducting_equipment}")

# Remark: Objects generated by network.objects(BaseType) are ordered by the name of their leaf
# class, so all `AcLineSegment`s will appear before all `Breaker`s.
# The `len_of` method returns the number of objects that inherit a specified type.
print(f"Number of switches: {network.len_of(Switch)}")
print(f"Number of conducting equipment: {network.len_of(ConductingEquipment)}")

print("\n#############\n# RESOLVERS #\n#############\n")
# There may be times when you need to reconstruct a network model from an unordered collection
# of identified objects. `NetworkService` allows you to add reference resolvers, which complete
# associations when the remaining object in an association is added.
network.resolve_or_defer_reference(per_length_impedance(line), "plsi_789")

print(f"Network has unresolved references? {network.has_unresolved_references()}")
print(f"plsi_789 has unresolved references? {network.has_unresolved_references('plsi_789')}")
for unresolved_reference in network.get_unresolved_references_from("acls_123"):
    print(f"Unresolved reference from acls_123: {unresolved_reference}")
for unresolved_reference in network.get_unresolved_references_to("plsi_789"):
    print(f"Unresolved reference to plsi_789: {unresolved_reference}")
print(f"Number of unresolved references to plsi_789: {network.num_unresolved_references('plsi_789')}")
print(f"Total unresolved references: {network.num_unresolved_references()}")
print("Adding plsi_789 to the network...")
network.add(PerLengthSequenceImpedance(mrid="plsi_789"))
print(f"Total unresolved references: {network.num_unresolved_references()}")
print(f"PerLengthSequenceImpedance of acls_123: {line.per_length_sequence_impedance}")

print("\n########################\n# CONNECTING TERMINALS #\n########################\n")
# Terminals in a `NetworkService` may be connected using the `connect_terminals` method.
# This automatically creates a connectivity node between the terminals, unless one of the
# terminals is already assigned to one.
t1, t2, t3 = (Terminal(mrid=f"t{i+1}") for i in range(3))
network.add(t1)
network.add(t2)
network.add(t3)
network.connect_terminals(t1, t2)
cn = t1.connectivity_node
print(f"Connected to node {cn}:")
for terminal in cn.terminals:
    print(f"\t{terminal}")

# The mrid of the connectivity node may also be used to connect a terminal
network.connect_by_mrid(t3, cn.mrid)
print(f"Connected to node {cn}:")
for terminal in cn.terminals:
    print(f"\t{terminal}")


print("\n#########\n# NAMES #\n#########\n")
# Apart from identified objects, a `NetworkService` also supports names. Each identified object has
# exactly one mRID, but can have any number of names. Each name has a name type. In this example,
# we add two names of type "NMI" to the network model.
meter1 = Meter()
meter2 = Meter()

name_type = NameType(name="NMI", description="National Meter Identifier")
name_type.get_or_add_name("987654321", line)
name_type.get_or_add_name("546372819", breaker)

network.add_name_type(name_type)
for name in network.get_name_type("NMI").names:
    print(f"NMI name {name.name} is assigned to {name.identified_object}")
for name_type in network.name_types:
    print(f"Network has name type {name_type}")

# Remark: In practice, NMI names are not assigned to lines and breakers.

print("\n####################\n# REMOVING OBJECTS #\n####################\n")

# You may use the `remove` method to remove objects from the network model.
network.remove(line)
print(f"{line} removed successfully.")
network.remove(breaker)
print(f"{breaker} removed successfully.")

# The object does not need to be the one that was added. It just needs to match the type and mRID.
plsi = PerLengthSequenceImpedance(mrid="plsi_789")
network.remove(plsi)
print(f"{plsi} removed successfully.")

# KeyError is raised if no matching object is found.
try:
    network.remove(line)
except KeyError as error:
    print(error)



#  Copyright 2025 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

# A Traversal is used to iterate over graph-like structures.
# The Evolve SDK contains several factory functions for traversals that cover common use cases.
import asyncio

from zepben.ewb import NetworkTraceStep, StepContext, Breaker, Switch, Feeder, LvFeeder, NetworkStateOperators

# For the purposes of this example, we will use the IEEE 13 node feeder.
from zepben.examples.ieee_13_node_test_feeder import network

feeder_head = network.get("br_650", Breaker)
switch = network.get("sw_671_692", Switch)
hv_feeder = network.get("hv_fdr", Feeder)
lv_feeder = network.get("lv_fdr", LvFeeder)


def reset_switch():
    switch.set_normally_open(False)
    switch.set_open(False)
    print("Switch reset (normally and currently closed)\n")


def print_heading(heading):
    print("+" + "-" * (len(heading) + 2) + "+")
    print(f"| {heading} |")
    print("+" + "-" * (len(heading) + 2) + "+\n")


async def network_trace():
    """
    Explanation of :class:`NetworkTrace` and its configurable options.

    More information about the constructors used in this function can be found in the docstring
    of :class:`NetworkTrace`
    """
    start_item = feeder_head

    async def network_traces():
        """
        :class:`NetworkTrace` iterates sequentially over all terminals in a network.

        By default, the trace will run:
            - Depth first, stepping to any equipment marked as 'in_service'.
            - Considering only the normal state of the network.
            - Performing :class:`StepAction`s only once per unique equipment encountered
        """
        from zepben.ewb import Tracing

        await (
            Tracing.network_trace()
        ).run(start_item)

    async def network_trace_state_operators():
        """
        Both the normal and current state of the network can be operated on by passing
        :class:`NetworkStateOperators` to the constructor as `network_state_operators`
        """
        from zepben.ewb import Tracing, NetworkStateOperators

        await (
            Tracing.network_trace(network_state_operators=NetworkStateOperators.NORMAL)
        ).run(start_item)

        await (
            Tracing.network_trace(network_state_operators=NetworkStateOperators.CURRENT)
        ).run(start_item)


    async def network_trace_queue():
        """
        :meth:`TraversalQueue.depth_first` or :meth:`TraversalQueue.breadth_first` can be passed in as
        `queue` to the constructor to control which step is taken next as the `NetworkTrace` traverses
        the network
        """
        from zepben.ewb import Tracing, TraversalQueue

        await (
            Tracing.network_trace(queue=TraversalQueue.depth_first())
        ).run(start_item)

        await (
            Tracing.network_trace(queue=TraversalQueue.breadth_first())
        ).run(start_item)

    async def network_trace_branching():
        """
        :class:`NetworkTrace` can also be configured to run in a branching manner. This intended to be
        used solely for tracing around loops both ways.

        A branching trace has the same defaults as a non_branching trace
        """

        from zepben.ewb import Tracing
        await (
            Tracing.network_trace_branching()
        ).run(start_item)

    # uncomment any of the following to not run them, as they have no StepActions they will do nothing
    # other than silently traverse the network.

    await network_traces()
    await network_trace_state_operators()
    await network_trace_queue()
    await network_trace_branching()


async def network_trace_step_actions():
    """
    Explanation of network trace :class:`StepActions` and :class:`NetworkTraceActionType`
    """
    start_item = feeder_head

    async def lambda_step_action():
        """
        A :class:`NetworkTrace` is useless as configured above, as we haven't specified any :class:`StepAction`s
        to take as we traverse. async functions are supported as step actions To get started, let's demonstrate
        a simple :class:`StepAction defined as a lambda.
        """
        from zepben.ewb import Tracing

        print_heading('NetworkTrace StepAction (as lambda):')
        await (
            Tracing.network_trace()
            .add_step_action(lambda step, _: print(step.path))
        ).run(start_item)

    async def function_step_action():
        """
        Functions can be used if you want type hinting, or more then one line.
        """
        from zepben.ewb import Tracing, NetworkTraceStep, StepContext

        print_heading('NetworkTrace StepAction (as function):')
        def print_step(step: NetworkTraceStep, context: StepContext) -> None:
            print(step.path)

        await (
            Tracing.network_trace()
            .add_step_action(print_step)
        ).run(start_item)

    async def subclassed_step_action():
        """
        And if it suits the need better, subclasses of :class:`StepAction` are also accepted, for this
        approach, please read the documentation of :class:`StepAction` as there are specific methods
        you will need to override.
        """
        from zepben.ewb import Tracing, NetworkTraceStep, StepAction, StepContext

        print_heading('NetworkTrace StepAction (as subclass):')
        class PrintingStepAction(StepAction):
            def __init__(self):
                super().__init__(self._apply)

            def _apply(self, step: NetworkTraceStep, context: StepContext):
                print(step.path)

        await (
            Tracing.network_trace()
            .add_step_action(PrintingStepAction())
        ).run(start_item)

    async def step_action_type():
        """"
        With :class:`StepAction`s you may wish to only execute these for every step taken, or once
        per equipment. This is configured by passing :class:`NetworkTraceActionType to the
        :class:`NetworkTrace` constructor.
        """
        from zepben.ewb import Tracing, NetworkTraceActionType

        print_heading('NetworkTrace (ALL_STEPS):')
        await (
            Tracing.network_trace(
                action_step_type=NetworkTraceActionType.ALL_STEPS
            )
            .add_step_action(lambda step, _: print(step.path))
        ).run(start_item)

        print_heading('NetworkTrace (FIRST_STEP_ON_EQUIPMENT):')
        await (
            Tracing.network_trace(
                action_step_type=NetworkTraceActionType.FIRST_STEP_ON_EQUIPMENT
            )
            .add_step_action(lambda step, _: print(step.path))
        ).run(start_item)

    # comment any of the following to skip running them

    await lambda_step_action()
    await function_step_action()
    await subclassed_step_action()
    await step_action_type()


async def network_trace_conditions():
    """
    Explanation of :class:`Conditions`
    """
    visited = list()

    start_item = feeder_head

    def print_step(ces: NetworkTraceStep, ctx: StepContext):
        visited.append(ces.path.to_equipment)
        print(f"\tDepth {ctx.step_number:02d}: {ces.path.to_equipment}")


    async def conditions_stop_at_open():
        """
        As :class:`NetworkTrace` will traverse all in service connected terminals regardless of open
        state, of we want to stop tracing at open switches etc. we need to add that as a condition.

        The condition is checked against the state specified with `network_state_operators` passed
        to the constructor of :class:`NetworkTrace`
        """
        from zepben.ewb import Tracing, stop_at_open, Switch

        print_heading("Network Trace Stopping at open equipment (NetworkStateOperators.NORMAL):")

        network.get("sw_671_692", Switch).set_normally_open(True)
        print("Switch set to normally open\n")

        await (
            Tracing.network_trace()
            .add_step_action(print_step)
            .add_condition(stop_at_open())
        ).run(start_item)

        print(f"Number of equipment visited: {len(visited)}")
        print()

        visited.clear()
        reset_switch()

    async def conditions_downstream():
        """
        You can specify a direction to trace to achieve a directed network trace.
        Tracing.set_direction() must be run on a network before performing any directed traces
        """
        from zepben.ewb import Tracing, downstream, upstream

        print_heading("Downstream Network Trace:")

        await Tracing.set_direction().run(network, network_state_operators=NetworkStateOperators.NORMAL)
        print("Feeder direction set for each terminal.\n")

        await (
            Tracing.network_trace()
            .add_step_action(print_step)
            .add_condition(downstream())
            #.add_condition(upstream())
        ).run(start_item)

        print(f"Number of equipment visited: {len(visited)}\n")

        await Tracing.clear_direction().run(start_item)
        visited.clear()

    async def conditions_limit_equipment_steps():
        """
        Limited connected equipment traces allow you to trace up to a number of steps.
        Running the trace returns a dictionary from each visited equipment to the number of steps
        away it is from a starting equipment.
        """
        from zepben.ewb import Tracing, downstream, limit_equipment_steps, AcLineSegment

        print_heading("Downstream NetworkTrace with limited equipment steps:")

        line = network.get("l_632_671", AcLineSegment)

        await Tracing.set_direction().run(network, network_state_operators=NetworkStateOperators.NORMAL)
        print("Feeder direction set for each terminal.\n")

        await (
            Tracing.network_trace()
            .add_condition(downstream())
            .add_stop_condition(limit_equipment_steps(limit=2))
            .add_step_action(print_step)
        ).run(line)

        await Tracing.clear_direction().run(start_item)
        print(f"Feeder direction removed for each terminal.")
        print()

        reset_switch()

    await conditions_stop_at_open()
    await conditions_downstream()
    await conditions_limit_equipment_steps()


async def assigning_equipment_to_feeders():
    """
    Use :meth:`assign_equipment_to_feeders` and :meth`assign_equipment_to_lv_feeders` to assign equipment to HV and LV feeders.

    :meth:`assign_equipment_to_feeders` also ensures that HV feeders that power LV feeders are associated.

    As with all tracing, both the normal and current state can be operated on by passing in :class:`NetworkStateOperators.NORMAL`
    or :class:`NetworkStateOperators.CURRENT`. e.g.

    .. code-block:: python

        Tracing.assign_equipment_to_feeders.run(
            network_state_operators=NetworkStateOperators.NORMAL
        )
    """
    from zepben.ewb import Tracing

    print_heading("ASSIGNING EQUIPMENT TO FEEDERS")

    print(f"Equipment in HV feeder: {[eq.mrid for eq in hv_feeder.equipment]}")
    print(f"Equipment in LV feeder: {[eq.mrid for eq in lv_feeder.equipment]}")
    print(f"LV feeders powered by HV feeder: {[lvf.mrid for lvf in hv_feeder.normal_energized_lv_feeders]}")
    print(f"HV feeders powering LV feeder: {[hvf.mrid for hvf in lv_feeder.normal_energizing_feeders]}")
    print()
    await Tracing.assign_equipment_to_feeders().run(network)
    await Tracing.assign_equipment_to_lv_feeders().run(network)
    print("Equipment assigned to feeders.")
    print()
    print(f"Equipment in HV feeder: {[eq.mrid for eq in hv_feeder.equipment]}")
    print(f"Equipment in LV feeder: {[eq.mrid for eq in lv_feeder.equipment]}")
    print(f"LV feeders powered by HV feeder: {[lvf.mrid for lvf in hv_feeder.normal_energized_lv_feeders]}")
    print(f"HV feeders powering LV feeder: {[hvf.mrid for hvf in lv_feeder.normal_energizing_feeders]}")
    print()


async def feeder_direction():
    """
    Examples on using set/clear direction to set or clear feeder directions to or from a network.
    """
    async def set_feeder_direction():
        """
        Use Tracing.set_direction().run(network) to set feeder directions to Terminals in the network.

        .. code-block:: python

            await Tracing.clear_direction().run(
                network
            )

        As with all tracing, both the normal and current state can be operated on by passing in :class:`NetworkStateOperators.NORMAL`
        or :class:`NetworkStateOperators.CURRENT`. e.g.

        .. code-block:: python

            await Tracing.clear_direction().run(
                network,
                network_state_operators=NetworkStateOperators.CURRENT
            )
        """
        from zepben.ewb import Tracing, NetworkStateOperators, Terminal

        print_heading("SETTING FEEDER DIRECTION")

        consumer_terminal = network.get("ec_675_t", Terminal)
        print(f"Normal feeder direction of HV feeder head terminal: {hv_feeder.normal_head_terminal.normal_feeder_direction}")
        print(f"Current feeder direction of HV feeder head terminal: {hv_feeder.normal_head_terminal.current_feeder_direction}")
        print(f"Normal feeder direction of energy consumer 675 terminal: {consumer_terminal.normal_feeder_direction}")
        print(f"Current feeder direction of energy consumer 675 terminal: {consumer_terminal.current_feeder_direction}")
        print()
        await Tracing.set_direction().run(network, network_state_operators=NetworkStateOperators.NORMAL)
        await Tracing.set_direction().run(network, network_state_operators=NetworkStateOperators.CURRENT)
        print("Normal and current feeder direction set.")
        print()
        print(f"Normal feeder direction of HV feeder head terminal: {hv_feeder.normal_head_terminal.normal_feeder_direction}")
        print(f"Current feeder direction of HV feeder head terminal: {hv_feeder.normal_head_terminal.current_feeder_direction}")
        print(f"Normal feeder direction of energy consumer 675 terminal: {consumer_terminal.normal_feeder_direction}")
        print(f"Current feeder direction of energy consumer 675 terminal: {consumer_terminal.current_feeder_direction}")
        print()

    async def clear_feeder_direction():
        """
        Use Tracing.clear_direction().run(network) to clear the feeder direction for Terminals in the network.

        .. code-block:: python

            await Tracing.set_direction().run(
                network
            )

        As with all tracing, both the normal and current state can be operated on by passing in
        :class:`NetworkStateOperators.NORMAL` or :class:`NetworkStateOperators.CURRENT`. e.g.

        .. code-block:: python

            await Tracing.set_direction().run(
                network,
                network_state_operators=NetworkStateOperators.CURRENT
            )
        """
        from zepben.ewb import Tracing, NetworkStateOperators, Terminal

        print_heading("REMOVING FEEDER DIRECTION")

        consumer_terminal = network.get("ec_675_t", Terminal)
        print(f"Normal feeder direction of HV feeder head terminal: {hv_feeder.normal_head_terminal.normal_feeder_direction}")
        print(f"Current feeder direction of HV feeder head terminal: {hv_feeder.normal_head_terminal.current_feeder_direction}")
        print(f"Normal feeder direction of energy consumer 675 terminal: {consumer_terminal.normal_feeder_direction}")
        print(f"Current feeder direction of energy consumer 675 terminal: {consumer_terminal.current_feeder_direction}")
        print()
        await Tracing.clear_direction().run(consumer_terminal, network_state_operators=NetworkStateOperators.NORMAL)
        await Tracing.clear_direction().run(consumer_terminal, network_state_operators=NetworkStateOperators.CURRENT)
        print("Normal and current feeder direction removed.")
        print()
        print(f"Normal feeder direction of HV feeder head terminal: {hv_feeder.normal_head_terminal.normal_feeder_direction}")
        print(f"Current feeder direction of HV feeder head terminal: {hv_feeder.normal_head_terminal.current_feeder_direction}")
        print(f"Normal feeder direction of energy consumer 675 terminal: {consumer_terminal.normal_feeder_direction}")
        print(f"Current feeder direction of energy consumer 675 terminal: {consumer_terminal.current_feeder_direction}")
        print()

        reset_switch()

    from zepben.ewb import SinglePhaseKind

    switch.set_normally_open(True, phase=SinglePhaseKind.A)
    print(f"Switch set to normally open on phase A. Switch is between feeder head and energy consumer 675.")

    await set_feeder_direction()
    await clear_feeder_direction()


async def trees():
    """
    A downstream tree contains all non-intersecting equipment paths starting from a common equipment
    and following downstream terminals. The same equipment may appear multiple times in the tree if
    the network contains multiple downstream paths to the equipment, i.e. loops. As this is backed by
    a NetworkTrace, either the normal or current state of the network may be used to determine whether
    to trace through each switch when combined with `Conditions.stop_at_open`
    """
    from zepben.ewb import Tracing, SinglePhaseKind, EquipmentTreeBuilder, TreeNode, NetworkStateOperators

    print_heading("DOWNSTREAM TREES")

    def desc_lines(node: TreeNode):
        children = list(node.children)
        for i, child in enumerate(children):
            is_last_child = i == len(children) - 1
            branch_char = "┗" if is_last_child else "┣"
            stem_char = " " if is_last_child else "┃"
            yield f"{branch_char}━{child.identified_object}"
            for line in desc_lines(child):
                yield f"{stem_char} {line}"

    def print_tree(root_node: TreeNode):
        print(root_node.identified_object)
        for line in desc_lines(root_node):
            print(line)
        print()

    switch.set_open(True, SinglePhaseKind.C)
    print("Switch set to currently open on phase C.\n")

    await Tracing.set_direction().run(network, network_state_operators=NetworkStateOperators.NORMAL)
    print("Feeder direction set.\n")

    print("Normal Downstream Tree:")
    equip_tree_builder = EquipmentTreeBuilder()
    await (
        Tracing.network_trace()
        .add_step_action(equip_tree_builder)
        .run(feeder_head)
    )
    print_tree(next(equip_tree_builder.roots))

    print("Current Downstream Tree:")
    cur_equip_tree_builder = EquipmentTreeBuilder()
    await (
        Tracing.network_trace(
            network_state_operators=NetworkStateOperators.CURRENT
        ).add_step_action(cur_equip_tree_builder)
    ).run(feeder_head)
    print_tree(next(cur_equip_tree_builder.roots))

    await Tracing.clear_direction().run(feeder_head)
    print(f"Feeder direction removed for each terminal.\n")

    reset_switch()


async def main():
    # All examples are self-contained. Feel free to comment out any of the following lines to isolate specific examples.
    await network_trace()
    await network_trace_step_actions()
    await network_trace_conditions()
    await assigning_equipment_to_feeders()
    await feeder_direction()
    await trees()

if __name__ == "__main__":
    asyncio.run(main())



#  Copyright 2025 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
import tempfile

from zepben.ewb import EwbDataFilePaths, DatabaseType
from pathlib import Path
from datetime import date

# This example can either create its own temporary directories or be use an existing EWB data path.
create_temp_files = True

if create_temp_files:
    print("Creating temporary database directories for example...")
    data_path = Path(tempfile.gettempdir())
    network_list = ["2012-09-10", "2021-06-14", "2023-06-13", "2023-07-10", "2023-09-09", "2023-11-10"]
    customer_list = ["2008-01-01", "2009-01-01", "2010-01-01", "2011-01-01", "2012-09-10"]
    for to_create in network_list:
        new_dir = data_path.joinpath(to_create)
        new_dir.mkdir(exist_ok=True)
        new_dir.joinpath(f"{to_create}-network-model.sqlite").touch()
    for to_create in customer_list:
        new_dir = data_path.joinpath(to_create)
        new_dir.mkdir(exist_ok=True)
        new_dir.joinpath(f"{to_create}-customers.sqlite").touch()
else:
    data_path = Path("</path/to/ewb/data/>")
    network_list = []
    customer_list = []

# Initialize EwbDataFilePaths with the EWB data directory
ewb_data = EwbDataFilePaths(data_path)

# List all the dates for which exist network databases in the data path
list_of_available_dates = ewb_data.get_network_model_databases()
print(f"\nAll network databases in data directory ({ewb_data.base_dir}):")
for available_date in list_of_available_dates:
    print(f"{available_date.isoformat()}")

# Find the first date for which exists a customer database before 2011-09-10
closest_date_before = ewb_data.find_closest(DatabaseType.CUSTOMER, target_date=date(2011, 9, 10))
print(f"\nThe last customer database before 2011-09-10: {closest_date_before.isoformat() or closest_date_before}")

if create_temp_files:
    if network_list and customer_list:
        for to_cleanup in network_list:
            date_dir = data_path.joinpath(to_cleanup)
            date_dir.joinpath(f"{to_cleanup}-network-model.sqlite").unlink()
        for to_cleanup in customer_list:
            date_dir = data_path.joinpath(to_cleanup)
            date_dir.joinpath(f"{to_cleanup}-customers.sqlite").unlink()
        for to_cleanup in customer_list + network_list:
            date_dir = data_path.joinpath(to_cleanup)
            if date_dir.exists():
                date_dir.rmdir()
        print("\nTemporary files successfully cleaned up.")
    else:
        print("\nUnexpected issue while attempting to cleanup temporary files.")



#  Copyright 2025 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

import os
from typing import List, Generator

import requests
from gql import gql, Client
from gql.transport.requests import RequestsHTTPTransport
from zepben.auth import get_token_fetcher

# This example utilises the EWB GraphQL APIs to fetch the network hierarchy from the server and
# then create a Powerfactory model by selecting components of the hierarchy to use.
# To use, populate the below variables with your desired targets plus the server and auth settings.

# Set of mRID/names of targets, leaving any target blank will exclude that level of hierarchy if
# it's the highest level. Names are visible through the hierarchy viewer in the UI - or you can do
# a getNetworkHierarchy GraphQL query as per the below.
target_zone_substation = {"zonesub-mRID-or-name"}
target_feeder = {"feeder-mRID-or-name"}
target_lv = {"lvfeeder-mRID-or-name"}

# resulting PFD file name
file_name = "test_file"
output_dir = "path to output dir"

# use feeder max demand for load?
# False will create time series characteristic for load
feeder_max_demand = False

# graphQL endpoint access settings
network_endpoint = 'https://{url}/api/network/graphql'
api_endpoint = 'https://{url}/api/graphql'
#Generate a personal access token from the energy work bench UI and replace <Token>.
token = 'Bearer <Token>'

### EXAMPLE QUERY ONLY ###
# This is an example GraphQL query for the full network hierarchy. This is not used as part of this
# code, and is purely illustrative.
# See below functions for actual queries used.
'''
query network {
    getNetworkHierarchy {
        geographicalRegions {
            mRID
            name
            subGeographicalRegions {
                mRID
                name
                substations {
                    mRID
                    name
                    feeders {
                        mRID
                        name
                        normalEnergizedLvFeeders {
                            mRID
                            name
                        }
                    }
                }
            }
        }
    }
}
'''
network_transport = RequestsHTTPTransport(url=network_endpoint, headers={'Authorization': token})
api_transport = RequestsHTTPTransport(url=api_endpoint, headers={'Authorization': token})

network_client = Client(transport=network_transport)
api_client = Client(transport=api_transport)

# full network hierarchy, simply remove levels that is not required
# Zone sub is the highest level supported in this example code
'''
query network {
    getNetworkHierarchy {
        geographicalRegions {
            mRID
            name
            subGeographicalRegions {
                mRID
                name
                substations {
                    mRID
                    name
                    feeders {
                        mRID
                        name
                        normalEnergizedLvFeeders {
                            mRID
                            name
                        }
                    }
                }
            }
        }
    }
}
'''


def request_pf_model_for_zone(graphql_body):
    """
    Request model for ZoneSub -> Feeder -> lvFeeder
    """
    if check_if_currently_generating_a_model():
        result = network_client.execute(graphql_body)  # retrieve network hierarchy
        target = list(get_target(result))
        model_id = request_pf_model(target, file_name, feeder_max_demand)
        print(f"Power factory model creation requested, model id: {model_id}")
    else:
        print("Warning: Still generating previous model, current model will not be generated.")


def get_target(result) -> Generator[str, None, None]:
    if target_zone_substation:
        for zone_sub in (x for x in result['getNetworkHierarchy']["substations"] if
                         x['mRID'] in target_zone_substation or x['name'] in target_zone_substation):
            yield zone_sub['mRID']
            yield from get_feeder(zone_sub)
    else:
        # No Zone sub was specified thus no zone sub will be added to target
        for zone_sub in result['getNetworkHierarchy']["substations"]:
            yield from get_feeder(zone_sub)


def get_feeder(zone_sub) -> Generator[str, None, None]:
    if 'feeders' in zone_sub.keys():
        if target_feeder:
            # Path to include only specific feeders
            for feeder in (x for x in zone_sub['feeders'] if x['mRID'] in target_feeder or x['name'] in target_feeder):
                yield feeder['mRID']
                yield from get_lvfeeder(feeder)
        else:
            for feeder in zone_sub['feeders']:
                if target_zone_substation:
                    yield feeder['mRID']
                yield from get_lvfeeder(feeder)


def get_lvfeeder(feeder) -> Generator[str, None, None]:
    if 'normalEnergizedLvFeeders' in feeder.keys():
        if target_lv:
            yield from (
                x['mRID'] for x in feeder['normalEnergizedLvFeeders']
                if x['mRID'] in target_lv or x['name'] in target_lv
            )
        else:
            # Path to include all lvFeeders
            yield from (lv['mRID'] for lv in feeder['normalEnergizedFeeders'])


def request_pf_model(equipment_container_list: List[str], filename: str, spread_max_demand: bool = False):
    """
    Performs the GraphQL request to create the Powerfactory model for the provided list of equipment containers.

    :param equipment_container_list: List of EquipmentContainer mRIDs to include in the Powerfactory model.
    :param filename: Desired PFD filename
    :param spread_max_demand: Whether to spread max demand load across transformers/loads. False will instead
                              configure the timeseries database.
    """
    # Set isPublic to false if you only want the specific user to see the model
    body = gql('''
    mutation createPowerFactoryModel($input: PowerFactoryModelInput!) {
        createPowerFactoryModel(input: $input)
    }
    ''')
    variables=dict(
        input=dict(
            name=filename,
            generationSpec=dict(
                equipmentContainerMrids=equipment_container_list,
                distributionTransformerConfig=dict(
                    rGround=0.01,
                    xGround=0.01
                ),
                loadConfig=dict(
                    spreadMaxDemand=spread_max_demand
                )
            ),
            isPublic='true'
        )
    )
    result = api_client.execute(body, variable_values=variables)
    return result['createPowerFactoryModel']


def check_if_currently_generating_a_model():
    body = gql('''
    query pagedPowerFactoryModels(
      $limit: Int!
      $offset: Long!
      $filter: GetPowerFactoryModelsFilterInput
      $sort: GetPowerFactoryModelsSortCriteriaInput
    ) {
        pagedPowerFactoryModels(
            limit: $limit
            offset: $offset
            filter: $filter
            sort: $sort
        ) {
            totalCount
            offset
            powerFactoryModels {
                id
                name
                createdAt
                state
                errors
            }
        }
    }
    ''')
    variables = dict(
        limit=10,
        offset=0,
        filter={}
    )
    result = api_client.execute(body, variable_values=variables)
    return not any(
        entry for entry in result['pagedPowerFactoryModels']['powerFactoryModels']
        if entry['state'] == 'CREATION'
    )


def download_model(model_number):
    # Request model
    model_url = api_endpoint.replace("graphql", "power-factory-model/") + str(model_number)
    body = gql('''
    query powerFactoryModelById($modelId: ID!) {
      powerFactoryModelById(modelId: $modelId) {
        id
        name
        createdAt
        state
        generationSpec {
          equipmentContainerMrids
          distributionTransformerConfig {
            rGround
            xGround
          }
        }
        isPublic
        errors
      }
    }
    ''')
    variables = dict(
        modelId=model_number
    )
    result = api_client.execute(body, variable_values=variables)
    model_status = result['powerFactoryModelById']['state']
    match model_status:
        case 'COMPLETED':
            model = requests.get(model_url, headers={'Authorization': token})
            open(os.path.join(output_dir, file_name) + ".pfd", 'wb').write(model.content)
            print(file_name + ".pfd saved at " + output_dir)
        case "CREATION":
            print("Model is still being created, please download at a later time")
        case "FAILED":
            print("Model creation error: " + str(result['powerFactoryModelById']['errors']))


graphql_queries = dict(
    zone_with_hv_lv=gql(
        '''
        query network {
          getNetworkHierarchy {
                substations {
                  mRID
                  name
                  feeders {
                    mRID
                    name
                    normalEnergizedLvFeeders {
                      mRID
                      name
                    }
                  }
                }
              }
            }
        '''
    ),
    zone_with_hv_only=gql(
        '''
        query network {
          getNetworkHierarchy {
                substations {
                  mRID
                  name
                  feeders {
                    mRID
                    name
                  }
                }
              }
            }
        '''
    )
)


if __name__ == "__main__":
    # Generate model with lv
    request_pf_model_for_zone(graphql_queries['zone_with_hv_lv'])

    # Generate model without lv
    request_pf_model_for_zone(graphql_queries['zone_with_hv_only'])

    # Download a model via model number
    download_model(123)



#  Copyright 2025 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Example trace showing method to traverse between any 2 given `IdentifiableObject` and build a list
of `ProtectedSwitch` objects found, if any
"""

import asyncio
import json
from typing import Tuple, Type

from zepben.ewb import (
    NetworkStateOperators, NetworkTraceActionType, NetworkTraceStep, StepContext, Tracing,
    NetworkConsumerClient, ProtectedSwitch, Recloser, LoadBreakSwitch, connect_with_token,
    IncludedEnergizedContainers
)


async def main(mrids: Tuple[str, str], io_type: Type[ProtectedSwitch], feeder_mrid):
    with open("config.json") as f:
        c = json.loads(f.read())

    channel = connect_with_token(
        host=c["host"],
        access_token=c["access_token"],
        rpc_port=c["rpc_port"]
    )
    client = NetworkConsumerClient(channel)
    await client.get_equipment_container(
        feeder_mrid,
        include_energized_containers=IncludedEnergizedContainers.LV_FEEDERS
    )
    network = client.service

    nodes = [network.get(_id) for _id in mrids]

    state_operators = NetworkStateOperators.NORMAL

    found_switch = set()
    found_node = []

    def stop_condition(step: NetworkTraceStep, context: StepContext):
        """if we encounter any of the equipment we have specified, we stop the trace and mark the `found_switch` list as valid"""
        if step.path.to_equipment in nodes:
            found_node.append(True)
            return True

    def step_action(step: NetworkTraceStep, context: StepContext):
        """Add any equipment matching the type passed in to the list, this list is invalid unless we trace onto our other node"""
        if isinstance(step.path.to_equipment, io_type):
            found_switch.add(step.path.to_equipment)

    trace = (
        Tracing.network_trace(
            network_state_operators=state_operators,
            action_step_type=NetworkTraceActionType.ALL_STEPS
        ).add_condition(state_operators.upstream())
        .add_stop_condition(stop_condition)
        .add_step_action(step_action)
    )

    queue = iter(nodes)
    while not found_node:  # run an upstream trace for every node specified until we encounter another specified node
        await trace.run(start=next(queue), can_stop_on_start_item=False)

    all(map(print, found_switch))  # print the list of switches
    print(bool(found_switch))  # print whether we found what we were looking for


if __name__ == "__main__":
    asyncio.run(main(mrids=('50735858', '66598892'), io_type=LoadBreakSwitch, feeder_mrid='RW1292'))
    asyncio.run(main(mrids=('50735858', '50295424'), io_type=Recloser, feeder_mrid='RW1292'))



#  Copyright 2025 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
from typing import Tuple

import numpy
from zepben.ewb import (
    AcLineSegment, Disconnector, PowerTransformer, TransformerFunctionKind, NetworkService, Terminal,
    PowerTransformerEnd, EnergyConsumer, PerLengthSequenceImpedance, PhaseCode, EnergyConsumerPhase,
    SinglePhaseKind, LinearShuntCompensator, ShuntCompensatorInfo, Feeder, LvFeeder, BaseVoltage, Breaker
)

__all__ = ["network"]

FEET_PER_MILE = 5280
METRES_PER_FOOT = 0.3048

#################
# BASE VOLTAGES #
#################
mv = BaseVoltage(mrid="mv", nominal_voltage=4_160)
lv = BaseVoltage(mrid="lv", nominal_voltage=480)

###########################
# EQUIPMENT AND TERMINALS #
###########################
# Terminal phases are ABC by default. The substation and its equipment is excluded from the model.
br_650_t1, br_650_t2 = Terminal(mrid="br_650_t1"), Terminal(mrid="br_650_t2")
br_650 = Breaker(mrid="br_650", terminals=[br_650_t1, br_650_t2], base_voltage=mv)

vr_650_632_t1, vr_650_632_t2 = Terminal(mrid="vr_650_632_t1"), Terminal(mrid="vr_650_632_t2") 
vr_650_632_e1 = PowerTransformerEnd(mrid="vr_650_632_e1", terminal=vr_650_632_t1, rated_u=mv.nominal_voltage)
vr_650_632_e2 = PowerTransformerEnd(mrid="vr_650_632_e2", terminal=vr_650_632_t2, rated_u=mv.nominal_voltage)
vr_650_632 = PowerTransformer(
    mrid="vr_650_632",
    function=TransformerFunctionKind.voltageRegulator,
    terminals=[vr_650_632_t1, vr_650_632_t2],
    power_transformer_ends=[vr_650_632_e1, vr_650_632_e2]
)

l_632_645_t1, l_632_645_t2 = (Terminal(mrid="l_632_645_t1", phases=PhaseCode.BCN),
                              Terminal(mrid="l_632_645_t2", phases=PhaseCode.BCN))
l_632_645 = AcLineSegment(
    mrid="l_632_645",
    length=500 * METRES_PER_FOOT,
    terminals=[l_632_645_t1, l_632_645_t2],
    base_voltage=mv
)

l_632_633_t1, l_632_633_t2 = (Terminal(mrid="l_632_633_t1", phases=PhaseCode.ABCN),
                              Terminal(mrid="l_632_633_t2", phases=PhaseCode.ABCN))
l_632_633 = AcLineSegment(
    mrid="l_632_633",
    length=500 * METRES_PER_FOOT,
    terminals=[l_632_633_t1, l_632_633_t2],
    base_voltage=mv
)

tx_633_634_t1, tx_633_634_t2 = (Terminal(mrid="tx_633_634_t1", phases=PhaseCode.ABCN),
                                Terminal(mrid="tx_633_634_t2", phases=PhaseCode.ABCN))
tx_633_634_e1 = PowerTransformerEnd(
    mrid="tx_633_634_e1",
    terminal=tx_633_634_t1,
    rated_u=mv.nominal_voltage
)
tx_633_634_e2 = PowerTransformerEnd(
    mrid="tx_633_634_e2",
    terminal=tx_633_634_t2,
    rated_u=lv.nominal_voltage
)
tx_633_634 = PowerTransformer(
    mrid="tx_633_634",
    terminals=[tx_633_634_t1, tx_633_634_t2],
    power_transformer_ends=[tx_633_634_e1, tx_633_634_e2]
)

l_645_646_t1, l_645_646_t2 = (Terminal(mrid="l_645_646_t1", phases=PhaseCode.BCN),
                              Terminal(mrid="l_645_646_t2", phases=PhaseCode.BCN))
l_645_646 = AcLineSegment(
    mrid="l_645_646",
    length=300 * METRES_PER_FOOT,
    terminals=[l_645_646_t1, l_645_646_t2],
    base_voltage=mv
)

l_650_632_t1, l_650_632_t2 = (Terminal(mrid="l_650_632_t1", phases=PhaseCode.ABCN),
                              Terminal(mrid="l_650_632_t2", phases=PhaseCode.ABCN))
l_650_632 = AcLineSegment(
    mrid="l_650_632",
    length=2000 * METRES_PER_FOOT,
    terminals=[l_650_632_t1, l_650_632_t2],
    base_voltage=mv
)

l_684_652_t1, l_684_652_t2 = (Terminal(mrid="l_684_652_t1", phases=PhaseCode.AN),
                              Terminal(mrid="l_684_652_t2", phases=PhaseCode.AN))
l_684_652 = AcLineSegment(
    mrid="l_684_652",
    length=800 * METRES_PER_FOOT,
    terminals=[l_684_652_t1, l_684_652_t2],
    base_voltage=mv
)

l_632_671_t1, l_632_671_t2 = (Terminal(mrid="l_632_671_t1", phases=PhaseCode.ABCN),
                              Terminal(mrid="l_632_671_t2", phases=PhaseCode.ABCN))
l_632_671 = AcLineSegment(
    mrid="l_632_671",
    length=2000 * METRES_PER_FOOT,
    terminals=[l_632_671_t1, l_632_671_t2],
    base_voltage=mv
)

l_671_684_t1, l_671_684_t2 = (Terminal(mrid="l_671_684_t1", phases=PhaseCode.ACN),
                              Terminal(mrid="l_671_684_t2", phases=PhaseCode.ACN))
l_671_684 = AcLineSegment(
    mrid="l_671_684",
    length=300 * METRES_PER_FOOT,
    terminals=[l_671_684_t1, l_671_684_t2],
    base_voltage=mv
)

l_671_680_t1, l_671_680_t2 = (Terminal(mrid="l_671_680_t1", phases=PhaseCode.ABCN),
                              Terminal(mrid="l_671_680_t2", phases=PhaseCode.ABCN))
l_671_680 = AcLineSegment(
    mrid="l_671_680",
    length=1000 * METRES_PER_FOOT,
    terminals=[l_671_680_t1, l_671_680_t2],
    base_voltage=mv
)

sw_671_692_t1, sw_671_692_t2 = (Terminal(mrid="sw_671_692_t1", phases=PhaseCode.ABCN),
                                Terminal(mrid="sw_671_692_t2", phases=PhaseCode.ABCN))
sw_671_692 = Disconnector(
    mrid="sw_671_692",
    terminals=[sw_671_692_t1, sw_671_692_t2],
    base_voltage=mv
)

l_684_611_t1, l_684_611_t2 = (Terminal(mrid="l_684_611_t1", phases=PhaseCode.CN),
                              Terminal(mrid="l_684_611_t2", phases=PhaseCode.CN))
l_684_611 = AcLineSegment(
    mrid="l_684_611",
    length=300 * METRES_PER_FOOT,
    terminals=[l_684_611_t1, l_684_611_t2],
    base_voltage=mv
)

l_692_675_t1, l_692_675_t2 = (Terminal(mrid="l_692_675_t1", phases=PhaseCode.ABCN),
                              Terminal(mrid="l_692_675_t2", phases=PhaseCode.ABCN))
l_692_675 = AcLineSegment(
    mrid="l_692_675",
    length=500 * METRES_PER_FOOT,
    terminals=[l_692_675_t1, l_692_675_t2],
    base_voltage=mv
)

ec_634_t = Terminal(mrid="ec_634_t", phases=PhaseCode.ABCN)
ec_634_pha = EnergyConsumerPhase(mrid="ec_634_pha", phase=SinglePhaseKind.A, p_fixed=160000, q_fixed=110000)
ec_634_phb = EnergyConsumerPhase(mrid="ec_634_phb", phase=SinglePhaseKind.B, p_fixed=120000, q_fixed=90000)
ec_634_phc = EnergyConsumerPhase(mrid="ec_634_phc", phase=SinglePhaseKind.C, p_fixed=120000, q_fixed=90000)
ec_634 = EnergyConsumer(
    mrid="ec_634",
    terminals=[ec_634_t],
    energy_consumer_phases=[ec_634_pha, ec_634_phb, ec_634_phc],
    base_voltage=lv
)

ec_645_t = Terminal(mrid="ec_645_t", phases=PhaseCode.ABCN)
ec_645_pha = EnergyConsumerPhase(mrid="ec_645_pha", phase=SinglePhaseKind.A, p_fixed=0, q_fixed=0)
ec_645_phb = EnergyConsumerPhase(mrid="ec_645_phb", phase=SinglePhaseKind.B, p_fixed=170000, q_fixed=125000)
ec_645_phc = EnergyConsumerPhase(mrid="ec_645_phc", phase=SinglePhaseKind.C, p_fixed=0, q_fixed=0)
ec_645 = EnergyConsumer(
    mrid="ec_645",
    terminals=[ec_645_t],
    energy_consumer_phases=[ec_645_pha, ec_645_phb, ec_645_phc],
    base_voltage=mv
)

ec_646_t = Terminal(mrid="ec_646_t", phases=PhaseCode.ABC)
ec_646_pha = EnergyConsumerPhase(mrid="ec_646_pha", phase=SinglePhaseKind.A, p=0, q=0)
ec_646_phb = EnergyConsumerPhase(mrid="ec_646_phb", phase=SinglePhaseKind.B, p=230000, q=132000)
ec_646_phc = EnergyConsumerPhase(mrid="ec_646_phc", phase=SinglePhaseKind.C, p=0, q=0)
ec_646 = EnergyConsumer(
    mrid="ec_646",
    terminals=[ec_646_t],
    energy_consumer_phases=[ec_646_pha, ec_646_phb, ec_646_phc],
    base_voltage=mv
)

ec_652_t = Terminal(mrid="ec_652_t", phases=PhaseCode.ABCN)
ec_652_pha = EnergyConsumerPhase(mrid="ec_652_pha", phase=SinglePhaseKind.A, p=128000, q=86000)
ec_652_phb = EnergyConsumerPhase(mrid="ec_652_phb", phase=SinglePhaseKind.B, p=0, q=0)
ec_652_phc = EnergyConsumerPhase(mrid="ec_652_phc", phase=SinglePhaseKind.C, p=0, q=0)
ec_652 = EnergyConsumer(
    mrid="ec_652",
    terminals=[ec_652_t],
    energy_consumer_phases=[ec_652_pha, ec_652_phb, ec_652_phc],
    base_voltage=mv
)

ec_671_t = Terminal(mrid="ec_671_t", phases=PhaseCode.ABC)
ec_671_pha = EnergyConsumerPhase(mrid="ec_671_pha", phase=SinglePhaseKind.A, p_fixed=385000, q_fixed=220000)
ec_671_phb = EnergyConsumerPhase(mrid="ec_671_phb", phase=SinglePhaseKind.B, p_fixed=385000, q_fixed=220000)
ec_671_phc = EnergyConsumerPhase(mrid="ec_671_phc", phase=SinglePhaseKind.C, p_fixed=385000, q_fixed=220000)
ec_671 = EnergyConsumer(
    mrid="ec_671",
    terminals=[ec_671_t],
    energy_consumer_phases=[ec_671_pha, ec_671_phb, ec_671_phc],
    base_voltage=mv
)

ec_675_t = Terminal(mrid="ec_675_t", phases=PhaseCode.ABCN)
ec_675_pha = EnergyConsumerPhase(mrid="ec_675_pha", phase=SinglePhaseKind.A, p_fixed=485000, q_fixed=190000)
ec_675_phb = EnergyConsumerPhase(mrid="ec_675_phb", phase=SinglePhaseKind.B, p_fixed=68000, q_fixed=60000)
ec_675_phc = EnergyConsumerPhase(mrid="ec_675_phc", phase=SinglePhaseKind.C, p_fixed=290000, q_fixed=212000)
ec_675 = EnergyConsumer(
    mrid="ec_675",
    terminals=[ec_675_t],
    energy_consumer_phases=[ec_675_pha, ec_675_phb, ec_675_phc],
    base_voltage=mv
)

ec_692_t = Terminal(mrid="ec_692_t", phases=PhaseCode.ABC)
ec_692_pha = EnergyConsumerPhase(mrid="ec_692_pha", phase=SinglePhaseKind.A, p=0, q=0)
ec_692_phb = EnergyConsumerPhase(mrid="ec_692_phb", phase=SinglePhaseKind.B, p=0, q=0)
ec_692_phc = EnergyConsumerPhase(mrid="ec_692_phc", phase=SinglePhaseKind.C, p=170000, q=151000)
ec_692 = EnergyConsumer(
    mrid="ec_692",
    terminals=[ec_692_t],
    energy_consumer_phases=[ec_692_pha, ec_692_phb, ec_692_phc],
    base_voltage=mv
)

ec_611_t = Terminal(mrid="ec_611_t", phases=PhaseCode.ABCN)
ec_611_pha = EnergyConsumerPhase(mrid="ec_611_pha", phase=SinglePhaseKind.A, p=0, q=0)
ec_611_phb = EnergyConsumerPhase(mrid="ec_611_phb", phase=SinglePhaseKind.B, p=0, q=0)
ec_611_phc = EnergyConsumerPhase(mrid="ec_611_phc", phase=SinglePhaseKind.C, p=170000, q=80000)
ec_611 = EnergyConsumer(
    mrid="ec_611",
    terminals=[ec_611_t],
    energy_consumer_phases=[ec_611_pha, ec_611_phb, ec_611_phc],
    base_voltage=mv
)

# Distributed load on line 632-671 is unmodelled.

lsc_675_t= Terminal(mrid="lsc_675_t1", phases=PhaseCode.ABCN)
lsc_675_info = ShuntCompensatorInfo(
    mrid="lsc_675_info",
    rated_voltage=4160,
    rated_current=48.077,
    rated_reactive_power=200000
)
lsc_675 = LinearShuntCompensator(mrid="lsc_675", terminals=[lsc_675_t], asset_info=lsc_675_info)

lsc_611_t = Terminal(mrid="lsc_611_t1", phases=PhaseCode.CN)
lsc_611_info = ShuntCompensatorInfo(
    mrid="lsc_611_info",
    rated_voltage=4160,
    rated_current=24.048,
    rated_reactive_power=100000
)
lsc_611 = LinearShuntCompensator(mrid="lsc_611", terminals=[lsc_611_t], asset_info=lsc_611_info)

###########################
# SETTING LINE IMPEDANCES #
###########################
def plsi_from_z_per_mile(mrid: str, *impedances: Tuple[float, float]):
    r_per_metre, x_per_metre = numpy.mean(impedances, axis=0) / (METRES_PER_FOOT * FEET_PER_MILE)
    return PerLengthSequenceImpedance(mrid=mrid, r=r_per_metre, x=x_per_metre)


plsi_601 = plsi_from_z_per_mile("plsi_601", (0.3465, 1.0179), (0.3375, 1.0478), (0.3414, 1.0348))
plsi_602 = plsi_from_z_per_mile("plsi_602", (0.7526, 1.1814), (0.7475, 1.1983), (0.7436, 1.2112))
plsi_603 = plsi_from_z_per_mile("plsi_603", (1.3294, 1.3471), (1.3238, 1.3569))
plsi_604 = plsi_from_z_per_mile("plsi_604", (1.3238, 1.3569), (1.3294, 1.3471))
plsi_605 = plsi_from_z_per_mile("plsi_605", (1.3292, 1.3475))
plsi_606 = plsi_from_z_per_mile("plsi_606", (0.7982, 0.4463), (0.7891, 0.4041), (0.7982, 0.4463))
plsi_607 = plsi_from_z_per_mile("plsi_607", (1.3425, 0.5124))

l_632_645.per_length_sequence_impedance = plsi_603
l_632_633.per_length_sequence_impedance = plsi_602
l_645_646.per_length_sequence_impedance = plsi_603
l_650_632.per_length_sequence_impedance = plsi_601
l_684_652.per_length_sequence_impedance = plsi_607
l_632_671.per_length_sequence_impedance = plsi_601
l_671_684.per_length_sequence_impedance = plsi_604
l_671_680.per_length_sequence_impedance = plsi_601
l_684_611.per_length_sequence_impedance = plsi_605
l_692_675.per_length_sequence_impedance = plsi_606

##############
# CONTAINERS #
##############
hv_fdr = Feeder(mrid="hv_fdr", normal_head_terminal=br_650_t2)
lv_fdr = LvFeeder(mrid="lv_fdr", normal_head_terminal=tx_633_634_t2)

##########################
# BUILDING NETWORK MODEL #
##########################
network = NetworkService()
for io in [br_650, br_650_t1, br_650_t2, vr_650_632, vr_650_632_t1, vr_650_632_t2, vr_650_632_e1, vr_650_632_e2, l_632_645, l_632_645_t1, l_632_645_t2,
           l_632_633, l_632_633_t1, l_632_633_t2, tx_633_634, tx_633_634_t1, tx_633_634_t2, tx_633_634_e1, tx_633_634_e2, l_645_646, l_645_646_t1, l_645_646_t2,
           l_650_632, l_650_632_t1, l_650_632_t2, l_684_652, l_684_652_t1, l_684_652_t2, l_632_671, l_632_671_t1, l_632_671_t2, l_671_684, l_671_684_t1,
           l_671_684_t2, l_671_680, l_671_680_t1, l_671_680_t2, sw_671_692, sw_671_692_t1, sw_671_692_t2, l_684_611, l_684_611_t1, l_684_611_t2, l_692_675,
           l_692_675_t1, l_692_675_t2, ec_634_t, ec_634_pha, ec_634_phb, ec_634_phc, ec_634, ec_645_t, ec_645_pha, ec_645_phb, ec_645_phc, ec_645, ec_646_t,
           ec_646_pha, ec_646_phb, ec_646_phc, ec_646, ec_652_t, ec_652_pha, ec_652_phb, ec_652_phc, ec_652, ec_671_t, ec_671_pha, ec_671_phb, ec_671_phc,
           ec_671, ec_675_t, ec_675_pha, ec_675_phb, ec_675_phc, ec_675, ec_692_t, ec_692_pha, ec_692_phb, ec_692_phc, ec_692, ec_611_t, ec_611_pha, ec_611_phb,
           ec_611_phc, ec_611, lsc_675_t, lsc_675_info, lsc_675, lsc_611_t, lsc_611_info, lsc_611, plsi_601, plsi_602, plsi_603, plsi_604, plsi_605, plsi_606,
           plsi_607, hv_fdr, lv_fdr, mv, lv]:
    network.add(io)

# Complete 650-632 regulator + line
network.connect_terminals(vr_650_632_t2, l_650_632_t1)

# Node 611
network.connect_terminals(l_684_611_t2, lsc_611_t)
network.connect_terminals(l_684_611_t2, ec_611_t)

# Node 632
network.connect_terminals(l_650_632_t2, l_632_633_t1)
network.connect_terminals(l_650_632_t2, l_632_645_t1)
network.connect_terminals(l_650_632_t2, l_632_671_t1)

# Node 633
network.connect_terminals(l_632_633_t2, tx_633_634_t1)

# Node 634
network.connect_terminals(tx_633_634_t2, ec_634_t)

# Node 645
network.connect_terminals(l_632_645_t2, l_645_646_t1)
network.connect_terminals(l_632_645_t2, ec_645_t)

# Node 646
network.connect_terminals(l_645_646_t2, ec_646_t)

# Node 650
network.connect_terminals(br_650_t2, vr_650_632_t1)

# Node 652
network.connect_terminals(l_684_652_t2, ec_652_t)

# Node 671
network.connect_terminals(l_632_671_t2, l_671_680_t1)
network.connect_terminals(l_632_671_t2, l_671_684_t1)
network.connect_terminals(l_632_671_t2, sw_671_692_t1)
network.connect_terminals(l_632_671_t2, ec_671_t)

# Node 675
network.connect_terminals(l_692_675_t2, lsc_675_t)
network.connect_terminals(l_692_675_t2, ec_675_t)

# Node 684
network.connect_terminals(l_671_684_t2, l_684_611_t1)
network.connect_terminals(l_671_684_t2, l_684_652_t1)

# Node 692
network.connect_terminals(sw_671_692_t2, l_692_675_t1)
network.connect_terminals(sw_671_692_t2, ec_692_t)



#  Copyright 2025 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
import asyncio
import json
from collections import defaultdict

from zepben.ewb import (
    Conductor, PowerTransformer, ConductingEquipment, EnergyConsumer, Switch, connect_with_token,
    NetworkConsumerClient, IncludedEnergizedContainers, IncludedEnergizingContainers
)


async def main():
    # See connecting_to_grpc_service.py for examples of each connect function
    print("Connecting to EWB..")
    with open("config.json") as f:
        c = json.loads(f.read())
    channel = connect_with_token(host=c["host"], access_token=c["access_token"], rpc_port=c["rpc_port"])
    feeder_mrid = "WD24"
    print(f"Fetching {feeder_mrid}")
    # Note you should create a new client for each Feeder you retrieve
    # There is also a NetworkConsumerClient that is asyncio compatible, with the same API.
    client = NetworkConsumerClient(channel=channel)
    network = client.service

    # Fetch feeder and all its LvFeeders
    await client.get_equipment_container(
        feeder_mrid,
        include_energized_containers=IncludedEnergizedContainers.LV_FEEDERS
    )

    print(f"\nTotal Number of objects: {client.service.len_of()}")
    # build a dictionary mapping of all objects returned organised by type.
    # e.g. {type(ConductingEquipment): [ce, ...]}
    objects_by_type = defaultdict(list)
    for ce in network.objects(ConductingEquipment):
        objects_by_type[type(ce)].append(ce)

    for t, objects in objects_by_type.items():
        print(f"Number of {t.__name__}'s = {len(objects)}")

    total_length = sum(c.length for c in network.objects(Conductor) if c.length)
    print(f"\nTotal conductor length in {feeder_mrid}: {total_length:.3f}m")

    feeder = network.get(feeder_mrid)
    print(f"\n{feeder.mrid} Transformers:")
    for eq in feeder.equipment:
        if isinstance(eq, PowerTransformer):
            print(f"    {eq} - Vector Group: {eq.vector_group.short_name}, Function: {eq.function.short_name}")

    print(f"\n\n{feeder_mrid} Energy Consumers:")
    for ec in network.objects(EnergyConsumer):
        print(f"    {ec} - Real power draw: {ec.q}W, Reactive power draw: {ec.p}VAr")

    print(f"\n{feeder_mrid} Switches:")
    for switch in network.objects(Switch):
        print(f"    {switch} - Open status: {switch.get_state():04b}")

    # === Some other examples of fetching containers ===

    # Fetch substation equipment and include equipment from HV/MV feeders powered by it
    await client.get_equipment_container(
        "substation ID",
        include_energized_containers=IncludedEnergizedContainers.FEEDERS
    )

    # Same as above, but also fetch equipment from LV feeders powered by the HV/MV feeders
    await client.get_equipment_container(
        "substation ID",
        include_energized_containers=IncludedEnergizedContainers.LV_FEEDERS
    )

    # Fetch feeder equipment without fetching any additional equipment from powering/powered containers
    await client.get_equipment_container("feeder ID")

    # Fetch HV/MV feeder equipment, the equipment from the substation powering it,
    # and the equipment from the LV feeders it powers
    await client.get_equipment_container(
        "feeder ID",
        include_energizing_containers=IncludedEnergizingContainers.SUBSTATIONS,
        include_energized_containers=IncludedEnergizedContainers.LV_FEEDERS
    )

    # Fetch LV feeder equipment and include equipment from HV/MV feeders powering it
    await client.get_equipment_container(
        "LV feeder ID",
        include_energizing_containers=IncludedEnergizingContainers.FEEDERS
    )

    # Same as above, but also fetch equipment from the substations powering the HV/MV feeders
    await client.get_equipment_container(
        "LV feeder ID",
        include_energizing_containers=IncludedEnergizingContainers.SUBSTATIONS
    )


if __name__ == "__main__":
    asyncio.run(main())



#  Copyright 2025 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

import asyncio
import csv
import json
import os
from typing import List, Union, Tuple, Optional, Dict

from zepben.ewb import NetworkConsumerClient, PhaseCode, AcLineSegment, FeederDirection, connect_with_token, \
    Tracing, downstream, NetworkTraceStep, ConductingEquipment, PowerTransformer, IncludedEnergizedContainers

LineInfo = Tuple[str, str, Optional[Union[int, float]]]


async def main():
    with open("config.json") as f:
        c = json.loads(f.read())

    print("Connecting to Server")
    channel = connect_with_token(host=c["host"], access_token=c["access_token"], rpc_port=c["rpc_port"])

    client = NetworkConsumerClient(channel)
    result = (await client.get_network_hierarchy()).throw_on_error().result
    print("Connection Established")

    tx_to_line_type: Dict[str, Tuple[List[LineInfo], bool]] = {}

    os.makedirs("csvs", exist_ok=True)
    for feeder in result.feeders.values():
        print(f"Fetching {feeder.mrid}")
        if not (network := await get_feeder_network(channel, feeder.mrid)):  # Skip feeders that fail to pull down
            print(f"Failed to retrieve feeder {feeder.mrid}")
            continue
        for io in network.objects(PowerTransformer):
            print(io)
            _loop = False

            for t in io.terminals:
                t_dir = t.normal_feeder_direction
                if t_dir == FeederDirection.BOTH:
                    _loop = True


            tx_to_line_type[io.mrid] = (await get_downstream_trace(io, PhaseCode.ABCN), _loop)
        await save_to_csv(tx_to_line_type, feeder.mrid)


async def save_to_csv(data: Dict[str, Tuple[List[LineInfo], bool]], feeder_mrid):
    filename = f"csvs/conductor_types_{feeder_mrid}.csv"
    with open(filename, mode='w', newline='') as file:
        writer = csv.writer(file)
        writer.writerow(["Feeder", "Transformer", "Line", "Line Type", "Length", "Loop"])

        for transformer, (values, loop) in data.items():
            for value in values:
                line, line_type, length = value
                switch_data = [feeder_mrid, transformer, line, line_type, length, loop]
                writer.writerow(switch_data)

    print(f"Data saved to {filename}")


async def get_feeder_network(channel, feeder_mrid):
    client = NetworkConsumerClient(channel)
    result = (
        await client.get_equipment_container(
            mrid=feeder_mrid,
            include_energized_containers=IncludedEnergizedContainers.LV_FEEDERS
        )
    )
    if result.was_failure:
        print(f"Failed: {result.thrown}")
        return None
    return client.service


async def get_downstream_trace(ce: ConductingEquipment, phase_code: PhaseCode) -> List[LineInfo]:
    l_type: List[LineInfo] = []

    def collect_eq_in(step: NetworkTraceStep, _):
        if isinstance(equip := step.path.to_equipment, AcLineSegment):
            nonlocal l_type
            l_type.append((equip.mrid, equip.asset_info.name, equip.length or 0))

    await (
        Tracing.network_trace()
        .add_condition(downstream())
        .add_step_action(collect_eq_in)
    ).run(start=ce, phases=phase_code)

    return l_type


if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())



#  Copyright 2025 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
import asyncio
import json
import os.path
from dataclasses import dataclass
import pandas as pd

from zepben.ewb import NetworkConsumerClient, connect_with_token, PowerTransformer

OUTPUT_FILE = "transformer_id_mapping.csv"
HEADER = True

with open("./config.json") as f:
    c = json.loads(f.read())


async def connect():
    channel = connect_with_token(host=c["host"], rpc_port=c["rpc_port"], access_token=c["access_token"], ca_filename=c["ca_path"])
    network_client = NetworkConsumerClient(channel=channel)

    if os.path.exists(OUTPUT_FILE):
        print(f"Output file {OUTPUT_FILE} already exists, please delete it if you would like to regenerate.")
        return

    network_hierarchy = (await network_client.get_network_hierarchy()).throw_on_error().value

    print("Network hierarchy:")
    for gr in network_hierarchy.geographical_regions.values():
        print(f"- Geographical region: {gr.name}")
        for sgr in gr.sub_geographical_regions:
            print(f"  - Subgeographical region: {sgr.name}")
            for sub in sgr.substations:
                print(f"    - Zone Substation: {sub.name}")
                await process_nodes(sub.mrid, channel)
                for fdr in sub.feeders:
                    print(f"      - Processing Feeder: {fdr.name}")
                    await process_nodes(fdr.mrid, channel)
                return  # Only process the first zone...


@dataclass
class NetworkObject(object):
    dist_tx_id: str
    dist_tx_name: str
    container: str
    container_mrid: str


async def process_nodes(container_mrid: str, channel):
    global HEADER
    print("Fetching from server ...")
    network_client = NetworkConsumerClient(channel=channel)
    network_service = network_client.service
    (await network_client.get_equipment_container(container_mrid)).throw_on_error()
    container = network_service.get(container_mrid)
    container_name = container.name

    print("Processing equipment ...")
    network_objects = []
    for equip in network_service.objects(PowerTransformer):
        no = NetworkObject(equip.mrid, equip.name, container_name, container_mrid)
        network_objects.append(no)

    network_objects = pd.DataFrame(network_objects)
    network_objects.to_csv(OUTPUT_FILE, index=False, mode='a', header=HEADER)
    print(f"Finished processing {container_mrid}")
    if HEADER:
        HEADER = False


if __name__ == "__main__":
    asyncio.run(connect())



#  Copyright 2025 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
import asyncio
import sys

from zepben.eas.client.feeder_load_analysis_input import FeederLoadAnalysisInput

from zepben.examples.utils import get_config_dir, get_client


async def main(argv):
    # See connecting_to_grpc_service.py for examples of each connect function
    config_dir = get_config_dir(argv)
    print("Connecting to EAS..")
    eas_client = get_client(config_dir)
    print("Connection established..")
    # Fire off a feeder load analysis study
    feeder_load_analysis_token = await eas_client.async_run_feeder_load_analysis_report(
        FeederLoadAnalysisInput(
            feeders=["feeder1", "feeder2"],
            substations=None,
            sub_geographical_regions=None,
            geographical_regions=None,
            start_date="2022-04-01",
            end_date="2022-12-31",
            fetch_lv_network=True,
            process_feeder_loads=True,
            process_coincident_loads=True,
            aggregate_at_feeder_level=False,
            output="Test"
        )
    )

    print(f"Feeder Load Analysis study: {feeder_load_analysis_token['data']['runFeederLoadAnalysis']}")

    # Feeder Load Analysis Study results can be retrieved from back end storage set up with EAS.

    await eas_client.aclose()


if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main(sys.argv))
